from typing import Optional

from .namedlinop import NamedLinop
from .nameddim import NS, Shape

__all__ = ["BreakpointLinop"]


class BreakpointLinop(NamedLinop):
    def __init__(self, ioshape: Optional[Shape] = None):
        super().__init__(NS(ioshape))

    def forward(self, x):
        return self.fn(self, x)

    @staticmethod
    def fn(linop, x, /):
        breakpoint()
        return x

    @staticmethod
    def adj_fn(linop, x, /):
        breakpoint()
        return x

    def split_forward(self, ibatch, obatch):
        return self

    def split_forward_fn(self, ibatch, obatch, /, data):
        return None



from typing import Optional
from torch import Tensor

import torch
import torch.nn as nn

from .namedlinop import NamedLinop
from .nameddim import NDorStr, ELLIPSES, NS, Shape
from torchlinops.utils import default_to
import torchlinops.functional as F
from torchlinops.functional._index.index import ensure_tensor_indexing


__all__ = ["Sampling"]


class Sampling(NamedLinop):
    """Sampling linop"""

    def __init__(
        self,
        idx: tuple,
        input_size: tuple,
        output_shape: Shape,
        input_shape: Optional[Shape] = None,
        batch_shape: Optional[Shape] = None,
    ):
        """
        Sampling: (B..., N...) -> (B..., M...)

        Parameters
        ----------
        idx : tuple of [M...] tensors
            One index for each "sampled" axis of the input tensor
            Use `canonicalize_idx` to turn a tensor of shape [M... D] to a D-tuple of index tensors.
            idx is in range [0, size-1]


        """
        self.input_size = input_size
        batch_shape = default_to(("...",), batch_shape)
        input_shape = default_to(("...",), input_shape)
        output_shape = default_to(("...",), output_shape)
        shape = NS(batch_shape) + NS(input_shape, output_shape)
        super().__init__(shape)
        self.register_shape("input_shape", input_shape)
        self.register_shape("output_shape", output_shape)
        self.register_shape("batch_shape", batch_shape)
        # if len(input_shape) != len(idx):
        #     raise ValueError(
        #         f"Input shape {input_shape} doesn't correspond to idx with shape {len(idx)}"
        #     )
        idx = ensure_tensor_indexing(idx, self.input_size)
        for d, (t, s) in enumerate(zip(idx, self.input_size)):
            if (t < 0).any() or (t >= s).any():
                raise ValueError(
                    f"Sampling index must lie within range [0, {s - 1}] but got range [{t.min().item()}, {t.max().item()}] for dim {d}"
                )
        self.idx = nn.ParameterList([nn.Parameter(i, requires_grad=False) for i in idx])

    @property
    def locs(self):
        """for compatibility with Interpolate linop"""
        return torch.stack(tuple(self.idx), dim=-1)

    @classmethod
    def from_mask(cls, mask, *args, **kwargs):
        """Alternative constructor for mask-based sampling"""
        idx = F.mask2idx(mask.bool())
        return cls(idx, *args, **kwargs)

    @classmethod
    def from_stacked_idx(cls, idx: Tensor, *args, dim=-1, **kwargs):
        """Alternative constructor for index in [M... D] form"""
        idx = F.canonicalize_idx(idx, dim=-1)
        return cls(idx, *args, **kwargs)

    def forward(self, x):
        return self.fn(self, x, tuple(self.idx))

    @staticmethod
    def fn(linop, x, idx):
        return F.index(x, idx)

    @staticmethod
    def adj_fn(linop, x, idx):
        return F.index_adjoint(x, idx, linop.input_size)

    def split_forward(self, ibatch, obatch):
        if self._shape.output_shape == ELLIPSES:
            # Cannot split if idx batch shape is not split
            return self
        return type(self)(
            self.split_forward_fn(ibatch, obatch, self.idx),
            self.input_size,
            self._shape.output_shape,
            self._shape.input_shape,
            self._shape.batch_shape,
        )

    def split_forward_fn(self, ibatch, obatch, idx):
        nM = len(idx[0].shape)
        if nM > 0:
            idx_slc = list(obatch[-nM:])
            return [i[idx_slc] for i in idx]
        return idx

    def register_shape(self, name, shape: tuple):
        self._shape.add(name, shape)

    def size(self, dim):
        if dim in self._shape.output_shape:
            dim_idx = self._shape.output_shape.index(dim)
            return self.locs.shape[dim_idx]
        elif dim in self._shape.input_shape:
            dim_idx = self._shape.input_shape.index(dim)
            return self.input_size[dim_idx]
        return None



from typing import Callable, Optional
from torch import Tensor

from dataclasses import dataclass

import torch
from tqdm import tqdm

from torchlinops.utils import default_to_dict, inner as zdot

__all__ = ["cg"]


def cg(
    A: Callable,
    y: Tensor,
    x0: Optional[Tensor] = None,
    max_num_iters: int = 20,
    gtol: float = 1e-3,
    ltol: float = 1e-5,
    disable_tracking: bool = False,
    tqdm_kwargs: Optional[dict] = None,
):
    """Solve Ax = y with conjugate gradients.

    A is a positive semidefinite matrix.



    Parameters
    ----------
    A : Callable
        Functional form of the (positive semidefinite) A matrix
    y : vector
        RHS of matrix equality
    x0 : Tensor, optional
        Initial guess at solution. If not provided, will initialize at zero
    max_num_iters : int, default 20
        The maximum number of iterations to run the algorithm
    tol : float, default 1e-6
        The relative change threshold for stopping early
    **tqdm_kwargs
        Extra keyword arguments to pass to tqdm

    Returns
    -------
    Tensor
        The result of conjugate gradient
    """
    # Default values
    if x0 is None:
        x = torch.zeros_like(y)
    else:
        x = x0.clone()
    tqdm_kwargs = default_to_dict(dict(desc="CG", leave=False), tqdm_kwargs)

    # Initialize run
    run = CGRun(ltol, gtol, A, y, disable=disable_tracking)
    run.update(x)

    r = y - A(x)
    p = r.clone()
    rs = zdot(r, r).real
    with tqdm(range(max_num_iters), **tqdm_kwargs) as pbar:
        for k in pbar:
            Ap = A(p)
            pAp = zdot(p, Ap)
            alpha = rs / pAp
            # Take step
            x = x + alpha * p
            r = r - alpha * Ap
            rs_old = rs.clone()
            rs = zdot(r, r).real
            run.update(x)
            # Stopping criterion
            if run.is_converged():
                break

            run.set_postfix(pbar)
            beta = rs / rs_old
            p = beta * p + r
    return run.x_out


@dataclass
class CGRun:
    """Track various aspects of the run

    Notes
    -----
    Assume A is positive definite.
    Letting A^(1/2) = B and z = B^(-H)y = B^(-1)y
    Then the solution to the least squares problem
        1/2||Bx - z||_2^2
    is given by the x satisfying
    B^HBx = B^Hz
    Ax = y <- The problem that CG solves

    For purposes of convergence testing, we can compute the gradients and loss values as
    grad = B^HBx - B^Hz
         = Ax - y
    loss = x^HB^HBx - x^HB^Hz - z^HBx + z^Hz
         = x^HAx - x^Hy - y^Hx + y^A^(-1)y



    """

    ltol: float
    gtol: float
    A: Callable
    y: Tensor
    x_out: Optional[Tensor] = None
    """The final tensor to output"""

    # Convergence
    prev_loss: float = None
    loss: float = float("inf")
    gnorm: float = None

    # Turn off tracking for speed
    disable: bool = False

    def update(self, x: Tensor):
        if self.disable:
            self.x_out = x
            return
        self.prev_loss = self.loss
        Ax = self.A(x)
        xy = zdot(x, self.y)
        self.loss = (zdot(x, Ax) - xy - xy.conj()).real.item()

        # Track best seen, or just update
        # if self.return_best:
        #     if self.loss < self.loss_best:
        #         self.x_out = x.clone()
        #         self.loss_best = self.loss
        # else:
        self.x_out = x

        # Compute grad norm
        grad = Ax - self.y
        self.gnorm = torch.linalg.vector_norm(grad).item()

    def set_postfix(self, pbar):
        if self.disable:
            return

        # Update progress bar if provided
        pbar.set_postfix(
            {
                "ldiff": abs(self.loss - self.prev_loss),
                "gnorm": self.gnorm,
            }
        )

    def is_converged(self) -> bool:
        if self.disable:
            return False
        ldiff = abs(self.loss - self.prev_loss)
        loss_converged = ldiff < self.ltol
        grad_converged = self.gnorm < self.gtol
        return loss_converged and grad_converged



from typing import Optional
from copy import copy
from torch import Tensor

from warnings import warn

import torch.nn as nn

import torchlinops.functional as F
from torchlinops.utils import default_to

from .namedlinop import NamedLinop
from .nameddim import ELLIPSES, NS, Shape

__all__ = ["ArrayToBlocks", "BlocksToArray"]


class ArrayToBlocks(NamedLinop):
    def __init__(
        self,
        grid_size: tuple[int, ...],
        block_size: tuple[int, ...],
        stride: tuple[int, ...],
        mask: Optional[Tensor] = None,
        batch_shape: Optional[Shape] = None,
        array_shape: Optional[Shape] = None,
        blocks_shape: Optional[Shape] = None,
    ):
        """

        mask : Tensor
        """
        self.grid_size = grid_size
        self.ndim = len(self.grid_size)
        self.block_size = block_size
        self.stride = stride

        self.batch_shape = default_to(("...",), batch_shape)
        self.array_shape = default_to(("...",), array_shape)
        self.blocks_shape = default_to(("...",), blocks_shape)
        shape = NS(self.batch_shape) + NS(self.array_shape, self.blocks_shape)
        super().__init__(shape)

        if mask is not None:
            self.mask = nn.Parameter(mask, requires_grad=False)
        else:
            self.mask = mask

    def forward(self, x):
        return self.fn(self, x)

    @staticmethod
    def fn(linop, x):
        return F.array_to_blocks(
            x,
            linop.block_size,
            linop.stride,
            linop.mask,
        )

    @staticmethod
    def adj_fn(linop, x):
        return F.blocks_to_array(
            x,
            linop.grid_size,
            linop.block_size,
            linop.stride,
            linop.mask,
        )

    @staticmethod
    def normal_fn(linop, x):
        return linop.adj_fn(linop, linop.fn(linop, x))

    def split_forward(self, ibatch, obatch):
        return copy(self)

    def adjoint(self):
        return BlocksToArray(
            self.grid_size,
            self.block_size,
            self.stride,
            self.mask,
            self.batch_shape,
            self.blocks_shape,
            self.array_shape,
        )

    def size(self, dim):
        return self.size_fn(dim)

    def size_fn(self, dim):
        ndim = len(self.grid_size)
        if dim in self.ishape[-ndim:]:
            i = self.ishape.index(dim) - len(self.ishape)
            return self.grid_size[i]
        return None


class BlocksToArray(NamedLinop):
    def __init__(
        self,
        grid_size: tuple[int, ...],
        block_size: tuple[int, ...],
        stride: tuple[int, ...],
        mask: Optional[Tensor] = None,
        batch_shape: Optional = None,
        blocks_shape: Optional = None,
        array_shape: Optional = None,
    ):
        self.grid_size = grid_size
        self.ndim = len(self.grid_size)
        self.block_size = block_size
        self.stride = stride

        self.batch_shape = default_to(("...",), batch_shape)
        self.blocks_shape = default_to(("...",), blocks_shape)
        self.array_shape = default_to(("...",), array_shape)
        shape = NS(self.batch_shape) + NS(self.blocks_shape, self.array_shape)
        super().__init__(shape)
        if mask is not None:
            self.mask = nn.Parameter(mask, requires_grad=False)
        else:
            self.mask = mask

    def forward(self, x):
        return self.fn(self, x)

    @staticmethod
    def fn(linop, x):
        return F.blocks_to_array(
            x,
            linop.grid_size,
            linop.block_size,
            linop.stride,
            linop.mask,
        )

    @staticmethod
    def adj_fn(linop, x):
        if x.shape[-linop.ndim :] != linop.grid_size:
            raise RuntimeError(
                f"BlocksToArray expected input with full size {linop.grid_size} but got {x.shape}"
            )
        return F.array_to_blocks(
            x,
            linop.block_size,
            linop.stride,
            linop.mask,
        )

    @staticmethod
    def normal_fn(linop, x):
        return linop.adj_fn(linop, linop.fn(linop, x))

    def split_forward(self, ibatch, obatch):
        return copy(self)

    def adjoint(self):
        return ArrayToBlocks(
            self.grid_size,
            self.block_size,
            self.stride,
            self.mask,
            self.batch_shape,
            self.array_shape,
            self.blocks_shape,
        )

    def size(self, dim):
        return self.size_fn(dim)

    def size_fn(self, dim):
        ndim = len(self.grid_size)
        if dim in self.oshape[-ndim:]:
            i = self.oshape.index(dim) - len(self.oshape)
            return self.grid_size[i]
        return None



import pytest
import torch


@pytest.fixture
def small1d():
    spec = {
        "N": (1,),
        "shape": (15,),
        "block_size": (3,),
        "stride": (1,),
        "mask": torch.tensor([0, 1, 0], dtype=bool),
    }
    return spec


@pytest.fixture
def medium1d():
    spec = {
        "N": (2, 2),
        "shape": (33,),
        "block_size": (7,),
        "stride": (2,),
    }
    return spec


@pytest.fixture
def large1d():
    spec = {
        "N": (20, 1),
        "shape": (1025,),
        "block_size": (25,),
        "stride": (25,),
    }
    return spec


@pytest.fixture
def tiny2d():
    spec = {
        "N": (1,),
        "shape": (4, 4),
        "block_size": (2, 2),
        "stride": (1, 1),
    }
    return spec


@pytest.fixture
def small2d():
    spec = {
        "N": (1,),
        "shape": (15, 15),
        "block_size": (3, 3),
        "stride": (1, 1),
        "mask": torch.tensor(
            [
                [0, 1, 0],
                [1, 1, 1],
                [0, 1, 0],
            ],
            dtype=bool,
        ),
    }
    return spec


@pytest.fixture
def medium2d():
    spec = {
        "N": (2, 2),
        "shape": (33, 43),
        "block_size": (7, 7),
        "stride": (2, 2),
    }
    return spec


@pytest.fixture
def large2d():
    spec = {
        "N": (20, 1),
        "shape": (125, 125),
        "block_size": (25, 25),
        "stride": (25, 25),
    }
    return spec


@pytest.fixture
def tiny3d():
    spec = {
        "N": (1,),
        "shape": (4, 4, 4),
        "block_size": (2, 2, 2),
        "stride": (1, 1, 1),
    }
    return spec


@pytest.fixture
def small3d():
    spec = {
        "N": (1,),
        "shape": (15, 15, 15),
        "block_size": (3, 3, 3),
        "stride": (1, 1, 1),
        "mask": torch.tensor(
            [
                [
                    [0, 0, 0],
                    [0, 1, 0],
                    [0, 0, 0],
                ],
                [
                    [0, 1, 0],
                    [1, 1, 1],
                    [0, 1, 0],
                ],
                [
                    [0, 0, 0],
                    [0, 1, 0],
                    [0, 0, 0],
                ],
            ],
            dtype=bool,
        ),
    }
    return spec


@pytest.fixture
def medium3d():
    spec = {
        "N": (2, 2),
        "shape": (33, 43, 23),
        "block_size": (7, 7, 5),
        "stride": (2, 2, 2),
    }
    return spec


@pytest.fixture
def large3d():
    spec = {
        "N": (1, 1),
        "shape": (192, 192, 192),
        "block_size": (128, 128, 128),
        "stride": (64, 64, 64),
    }
    return spec


@pytest.fixture
def verylarge3d():
    spec = {
        "N": (1, 1),
        "shape": (256, 256, 256),
        "block_size": (192, 192, 192),
        "stride": (64, 64, 64),
    }
    return spec



import pytest
from collections import OrderedDict

from torchlinops import NamedShape, NS, ND


def test_getter_setter():
    shape = NS(("A", "B"), ("C",))
    assert shape.ishape == ("A", "B")
    assert shape.oshape == ("C",)


def test_new_batch():
    shape = NS(("A", "B"), ("C",))
    shape.add("shared_batch", ("A",))
    assert shape.shared_batch == ("A",)

    shape.ishape = ("E", "F")
    assert shape.shared_batch == ("E",)


def test_new_dict():
    shape = NS(("A", "B"), ("C",))
    shape.add("axes_lengths", {"A": 2})
    assert shape.axes_lengths[ND.infer("A")] == 2

    shape.ishape = ("E", "F")
    assert shape.axes_lengths[ND.infer("E")] == 2


def test_random_attribute():
    shape = NS(("A", "B", ("C",)))
    # This should be ok
    shape.im_size = (64, 64, 64)


def test_diag():
    shape = NS(("A", "B"))
    assert shape.ishape == ("A", "B")
    assert shape.oshape == ("A", "B")
    shape.ishape = ("C", "D")
    assert shape.oshape == ("C", "D")


def test_empty_dim():
    shape = NS(tuple(), ("C",))
    assert shape.oshape == ("C",)


# @pytest.mark.skip
# def test_adjoint():
#     shape = NamedShape(("A", "B"), ("C",))
#     adj_shape = shape.H
#     adj_shape.ishape = ("D",)
#     assert shape.oshape == ("D",)


def test_product():
    shape1 = NS(("A", "B"), ("C",))
    shape2 = NS(("E", "F"))

    shape12 = shape1 + shape2
    assert shape12.ishape == ("A", "B", "E", "F")
    assert shape12.oshape == ("C", "E", "F")

    shape21 = shape2 + shape1
    assert shape21.ishape == ("E", "F", "A", "B")
    assert shape21.oshape == ("E", "F", "C")


def test_ellipses():
    shape1 = NS(("...", "A"))
    shape2 = NS(("C", "A"))
    assert shape1 == shape2

    shape1 = NS(("A", "B"))
    shape2 = NS(("C", "A"))
    assert shape1 != shape2

    shape1 = NS(("...", "C"), ("...", "D"))
    shape2 = NS(("...",))
    assert shape1 == shape2



__all__ = ["default_to", "default_to_dict"]


def default_to(*vals, typecast: bool = False):
    """Get the first non-None value, right to left order.

    Most "default" value goes first.
    """
    if len(vals) == 0:
        return None
    typecls = type(vals[0])
    if len(vals) == 1:
        return vals[0]
    for val in reversed(vals):
        if val is not None:
            if typecast:
                return typecls(val)
            return val


def default_to_dict(*dicts) -> dict:
    """Update a dict with default values progressively from left to right."""
    if len(dicts) == 0:
        return {}
    out = {}
    for d in dicts:
        if d is None:
            pass
        elif not isinstance(d, dict):
            raise ValueError(
                f"Non-dictionary found during default dictionary creation: {d}"
            )
        else:
            out.update(d)
    return out



import pytest
from math import prod

import torch
import sigpy as sp
import numpy as np

from torchlinops import NUFFT, PadLast, Interpolate
from torchlinops.functional._interp.tests._valid_pts import get_valid_locs
from torchlinops.tests.test_base import BaseNamedLinopTests
from torchlinops.utils import cifftn

from torchlinops.functional import nufft, nufft_adjoint


class TestNUFFT(BaseNamedLinopTests):
    equality_check = "approx"

    oversamp = [1.0, 1.25]

    instances = ["small3d"]

    # Unstable numerical behavior
    isclose_kwargs: dict = {"rtol": 1e-3}

    @pytest.fixture(scope="class", params=instances)
    def linop_input_output(self, request):
        spec = request.param
        spec = request.getfixturevalue(spec)
        width = spec["width"]
        oversamp = spec["oversamp"]
        grid_size = spec["grid_size"]
        locs_batch_size = spec["locs_batch_size"]
        ndim = len(grid_size)
        npts = prod(locs_batch_size)
        batch_size = spec["N"]
        ishape = (*batch_size, *grid_size)
        oshape = (*batch_size, *locs_batch_size)
        locs = get_valid_locs(
            locs_batch_size, grid_size, ndim, width, "cpu", centered=True
        )

        linop = NUFFT(
            locs.clone(),
            grid_size,
            output_shape=("R", "K"),
            width=width,
            oversamp=oversamp,
        )
        # Limit randomness
        x = 0.5 * torch.rand(ishape, dtype=torch.complex64, device="cpu") + 1
        y = 0.5 * torch.rand(oshape, dtype=torch.complex64, device="cpu") + 1
        y /= torch.linalg.vector_norm(locs, dim=-1)

        # Save original locs
        linop._locs_orig = locs

        return linop, x, y

    @pytest.fixture(scope="class")
    def small3d(self, request):
        N = (2, 1)
        # grid_size = (16, 16, 24)
        grid_size = (32, 32, 32)
        locs_batch_size = (3, 5)
        width = 4.0
        oversamp = 1.25

        spec = {
            "N": N,
            "grid_size": grid_size,
            "locs_batch_size": locs_batch_size,
            "width": width,
            "oversamp": oversamp,
        }
        return spec

    def test_nufft_sigpy(self, linop_input_output):
        A, x, y = linop_input_output
        coord = A._locs_orig.numpy()  # Not usually a param, only here for testing
        # sz = np.array(A.grid_size)
        # coord = np.where(coord <= (sz / 2), coord, coord - sz)
        width = A.width
        oversamp = A.oversamp

        Ax = A(x).numpy()
        Ax_sp = sp.nufft(x.numpy(), coord, oversamp=oversamp, width=width)
        assert np.allclose(Ax, Ax_sp, **self.isclose_kwargs)

        AHy = A.H(y).numpy()
        AHy_sp = sp.nufft_adjoint(
            y.numpy(), coord, x.shape, oversamp=oversamp, width=width
        )
        assert np.allclose(AHy, AHy_sp, **self.isclose_kwargs)

    def test_nufft_sigpy_functional(self, linop_input_output):
        A, x, y = linop_input_output
        coord = A._locs_orig.numpy()  # Not usually a param, only here for testing
        # sz = np.array(A.grid_size)
        # coord = np.where(coord <= (sz / 2), coord, coord - sz)
        width = A.width
        oversamp = A.oversamp

        Ax = nufft(x, A._locs_orig, oversamp, width).numpy()
        Ax_sp = sp.nufft(x.numpy(), coord, oversamp=oversamp, width=width)
        assert np.allclose(Ax, Ax_sp, **self.isclose_kwargs)

        ndim = A._locs_orig.shape[-1]
        grid_size = x.shape[-ndim:]
        AHy = nufft_adjoint(y, A._locs_orig, grid_size, oversamp, width).numpy()
        AHy_sp = sp.nufft_adjoint(
            y.numpy(), coord, x.shape, oversamp=oversamp, width=width
        )
        assert np.allclose(AHy, AHy_sp, **self.isclose_kwargs)


@pytest.fixture
def nufft_params():
    width = 6.0
    oversamp = 2.0
    grid_size = (120, 119, 146)
    padded_size = [int(i * oversamp) for i in grid_size]
    locs = get_valid_locs((10,), grid_size, len(grid_size), width, "cpu", centered=True)
    return {
        "width": width,
        "oversamp": oversamp,
        "grid_size": grid_size,
        "padded_size": padded_size,
        "locs": locs,
    }


def test_apodize(nufft_params):
    width = nufft_params["width"]
    oversamp = nufft_params["oversamp"]
    grid_size = nufft_params["grid_size"]
    padded_size = nufft_params["padded_size"]

    beta = NUFFT.beta(width, oversamp)
    apod = NUFFT.apodize_weights(grid_size, padded_size, oversamp, width, beta)

    x = np.ones(grid_size)
    apod_sp = sp.fourier._apodize(x, len(grid_size), oversamp, width, beta)
    assert np.allclose(apod, apod_sp)


def test_nufft_os_pad(nufft_params):
    grid_size = nufft_params["grid_size"]
    padded_size = nufft_params["padded_size"]
    pad = PadLast(padded_size, grid_size)

    x = torch.randn(*grid_size)
    padx = pad(x).numpy()

    padx_sp = sp.util.resize(x.numpy(), padded_size)

    assert np.allclose(padx, padx_sp)


def test_scale_locs(nufft_params):
    grid_size = nufft_params["grid_size"]
    padded_size = nufft_params["padded_size"]
    oversamp = nufft_params["oversamp"]

    # Torch version
    locs = nufft_params["locs"]
    locs_scaled = NUFFT.prep_locs(locs.clone(), grid_size, padded_size)

    coord = locs.clone().numpy()
    sz = np.array(grid_size)
    coord_scaled = sp.fourier._scale_coord(coord, grid_size, oversamp)
    assert np.allclose(locs_scaled, coord_scaled)


def test_nufft_interp(nufft_params):
    locs = nufft_params["locs"]
    grid_size = nufft_params["grid_size"]
    padded_size = nufft_params["padded_size"]
    width = nufft_params["width"]
    oversamp = nufft_params["oversamp"]
    beta = NUFFT.beta(width, oversamp)

    locs_prepared = NUFFT.prep_locs(locs.clone(), grid_size, padded_size)
    interp = Interpolate(
        locs_prepared,
        padded_size,
        batch_shape=None,
        locs_batch_shape=None,
        grid_shape=None,
        width=width,
        kernel="kaiser_bessel",
        kernel_params=dict(beta=beta),
    )

    x = torch.randn(*padded_size, dtype=torch.complex64)
    interpx = interp(x)

    interpx_sp = sp.interp.interpolate(
        x.numpy(),
        locs_prepared.numpy(),
        kernel="kaiser_bessel",
        width=width,
        param=beta,
    )

    assert np.allclose(interpx, interpx_sp, rtol=1e-3)


@pytest.mark.gpu
@pytest.mark.skipif(
    not torch.cuda.is_available(), reason="GPU is required but not available"
)
def test_nufft_device(nufft_params):
    locs = nufft_params["locs"]
    grid_size = nufft_params["grid_size"]
    width = nufft_params["width"]
    oversamp = nufft_params["oversamp"]
    linop = NUFFT(
        locs.clone(),
        grid_size,
        output_shape=("K",),
        width=width,
        oversamp=oversamp,
    )
    assert linop.device.type == "cpu"
    linop.to(torch.device("cuda"))
    assert linop.device.type == "cuda"



"""Functions for interoperability between cupy and pytorch."""

from torch import Tensor

from warnings import warn
import torch
import numpy as np

try:
    import cupy as cp

    cupy_enabled = True
except ImportError as e:
    warn("Cupy not available")
    import numpy as cp

    cupy_enabled = False


__all__ = ["to_pytorch", "from_pytorch", "get_device"]


def get_device(arr) -> torch.device:
    """Bootleg version of sigpy's more-robust get_device"""
    if isinstance(arr, np.ndarray):
        return torch.device("cpu")
    elif cupy_enabled and isinstance(arr, cp.ndarray):
        return torch.device(f"cuda:{arr.device.id}")
    elif isinstance(arr, Tensor):
        return arr.device
    else:
        raise ValueError(f"Can't get device of array type: {type(arr)}")


def to_pytorch(array, requires_grad: bool = False):
    """Zero-copy conversion from numpy/cupy array to pytorch tensor.

    For complex array input, returns a tensor with shape + [2],
    where tensor[..., 0] and tensor[..., 1] represent the real
    and imaginary.

    Args:
        array (numpy/cupy array): input.
        requires_grad(bool): Set .requires_grad output tensor
    Returns:
        PyTorch tensor.

    """
    import torch
    from torch.utils.dlpack import from_dlpack

    device = get_device(array)

    if device.type == "cpu":
        tensor = torch.from_numpy(array)
    else:
        tensor = from_dlpack(array.toDlpack())

    tensor.requires_grad = requires_grad
    return tensor.contiguous()


def from_pytorch(tensor):  # pragma: no cover
    """Zero-copy conversion from pytorch tensor to numpy/cupy array.

    If iscomplex, then tensor must have the last dimension as 2,
    and the output will be viewed as a complex valued array.

    Args:
        tensor (PyTorch tensor): input.
        iscomplex (bool): whether input represents complex valued tensor.

    Returns:
        Numpy/cupy array.

    """
    from torch.utils.dlpack import to_dlpack

    device = tensor.device
    if device.type == "cpu":
        output = tensor.detach().contiguous().numpy()
    else:
        if cupy_enabled:
            output = cp.from_dlpack(to_dlpack(tensor.contiguous()))
        else:
            raise TypeError(
                "CuPy not installed, " "but trying to convert GPU PyTorch Tensor."
            )

    # No longer necessary
    # if iscomplex:
    #     if output.shape[-1] != 2:
    #         raise ValueError('shape[-1] must be 2 when iscomplex is '
    #                          'specified, but got {}'.format(output.shape))

    #     with backend.get_device(output):
    #         if output.dtype == np.float32:
    #             output = output.view(np.complex64)
    #         elif output.dtype == np.float64:
    #             output = output.view(np.complex128)

    #         output = output.reshape(output.shape[:-1])

    return output


def to_pytorch_function(linop):  # pragma: no cover
    """Convert SigPy Linop to PyTorch Function.

    The returned function can be treated as a native
    pytorch function performing the linop operator.
    The function can be backpropagated, applied on GPU arrays,
    and has minimal overhead as the underlying arrays
    are shared without copying.
    For complex valued input/output, the appropriate options
    should be set when calling the function.

    Args:
        linop (Linop): linear operator to be converted.
        input_iscomplex (bool): whether the PyTorch input
            represents complex tensor.
        output_iscomplex (bool): whether the PyTorch output
            represents complex tensor.

    Returns:
        torch.autograd.Function: equivalent PyTorch Function.

    """
    import torch

    class LinopFunction(torch.autograd.Function):
        @staticmethod
        def forward(input):
            return to_pytorch(linop(from_pytorch(input)), input.requires_grad)

        @staticmethod
        def setup_context(ctx, inputs, output):
            pass

        @staticmethod
        def backward(ctx, grad_output):
            return to_pytorch(linop.H(from_pytorch(grad_output)))

        to_nested_mapping = linop

    return LinopFunction



import gc
import itertools
import logging
from collections import defaultdict
from copy import deepcopy
from typing import Literal, Optional, TypeVar
from warnings import warn

import torch
import torch.nn as nn
from torch import Tensor

__all__ = [
    "cdata",
    "tensor_memory_span",
    "get_device",
    "device_ordinal",
    "same_storage",
    "MemReporter",
    "memory_aware_to",
    "memory_aware_deepcopy",
]

T = TypeVar("T")
logger = logging.getLogger("torchlinops.utils")


def memory_aware_to(
    module: nn.Module,
    device: Optional[torch.device] = None,
    storage_map: Optional[dict] = None,
):
    """Move a module to a device, without unnecessary memory overhead."""
    if storage_map is None:
        storage_map = create_shared_buffer_map(module, device, copy=False)

    # Remember which modules were visited
    memo = set()

    def remap(m, level=0):
        if id(m) in memo:
            return
        logger.debug("\t" * level + f"{type(m).__name__}")
        for name, t in m._parameters.items():
            if t is not None and t.device != device:
                new_t = as_view_on_moved(t, storage_map)
                m._parameters[name] = nn.Parameter(new_t, requires_grad=t.requires_grad)
        for name, t in m._buffers.items():
            if t is not None and t.device != device:
                new_t = as_view_on_moved(t, storage_map)
                m._buffers[name] = new_t
        for child in m.children():
            remap(child, level + 1)
        memo.add(id(m))

    remap(module)
    return module


def memory_aware_deepcopy(module):
    """Deepcopy a module, without unnecessary memory overhead."""
    storage_map = create_shared_buffer_map(module, copy=True)

    # Recursively
    def copy_memory_aware(m: nn.Module):
        cls = type(m)
        new = cls.__new__(cls)
        new.__dict__ = m.__dict__.copy()
        new._parameters = dict()
        new._buffers = dict()
        new._modules = dict()

        for name, t in m._parameters.items():
            if t is not None:
                new_t = as_view_on_moved(t, storage_map)
                new._parameters[name] = nn.Parameter(
                    new_t, requires_grad=t.requires_grad
                )
        for name, t in m._buffers.items():
            if t is not None:
                new_t = as_view_on_moved(t, storage_map)
                new._buffers[name] = new_t
        for module_name, child_module in m._modules.items():
            new._modules[module_name] = copy_memory_aware(child_module)
        return new

    new_module = copy_memory_aware(module)
    return new_module


def create_shared_buffer_map(module, device=None, copy=False) -> dict:
    """Construct the smallest set of tensors that can be used to hold all the parameters in a module."""
    storage_tensors = defaultdict(list)

    def collect(m):
        """Recursively collect all parameters and buffers in the module."""
        for name, t in list(m.named_parameters(recurse=False)) + list(
            m.named_buffers(recurse=False)
        ):
            if t is None:
                continue
            key = cdata(t)
            storage_tensors[key].append(t)
        for child in m.children():
            collect(child)

    collect(module)

    # Maps from cdata to the buffer on target device that the tensor should go to.
    storage_map = {}
    for key, tensors in storage_tensors.items():
        if all(tensor.device == device for tensor in tensors) and not copy:
            # No new memory need be allocated for this block.
            continue
        # Find max offset + size for any tensor sharing this storage
        min_offset = float("inf")
        max_offset = float("-inf")
        dtype = None
        device_orig = None
        for t in tensors:
            new_min_offset, new_max_offset = tensor_memory_span(t)
            # size_bytes = t.element_size() * max_storage_size(t)
            # offset_bytes = t.storage_offset() * t.element_size()
            min_offset = min(min_offset, new_min_offset)
            max_offset = max(max_offset, new_max_offset)
            dtype = t.dtype
            device_orig = t.device

        # Create a flat tensor that spans the full required storage
        base_tensor = torch.empty(
            max_offset - min_offset + 1, dtype=dtype, device=device_orig
        )
        base_device = device if device is not None else device_orig
        storage_map[key] = (base_tensor.to(base_device), min_offset)
    return storage_map


def tensor_memory_span(t: torch.Tensor):
    """Compute the min and max memory offsets of a tensor.

    Supports negative strides (even though PyTorch doesn't.)

    """
    shape = t.shape
    strides = t.stride()
    offset = t.storage_offset()

    # For each dimension, pick the index that causes min and max access
    # If stride >= 0: (0, size-1), else (size-1, 0)
    index_ranges = [
        (0, s - 1) if stride >= 0 else (s - 1, 0) for s, stride in zip(shape, strides)
    ]

    # Generate all corners
    corners = itertools.product(*index_ranges)

    # Compute linear memory offset for each corner
    mem_offsets = [offset + sum(i * s for i, s in zip(idx, strides)) for idx in corners]

    return min(mem_offsets), max(mem_offsets)


def cdata(t: Tensor):
    """Get a pointer to the block of memory that the tensor is part of."""
    return t.untyped_storage()._cdata


def max_storage_size(tensor):
    """Compute the number of elements spanned by a strided tensor."""
    if tensor.numel() == 0:
        return 0
    size = 0
    for i, (dim, stride) in enumerate(zip(tensor.size(), tensor.stride())):
        size += (dim - 1) * stride
    return size + 1


def as_view_on_moved(tensor, storage_map):
    """Copy the data from a tensor to the actual target location in the target allocated storage."""
    key = cdata(tensor)
    base, min_offset = storage_map[key]
    view = base.as_strided(
        tensor.size(), tensor.stride(), tensor.storage_offset() - min_offset
    ).copy_(tensor)
    return view


def get_device(device_idx: int = -1):
    return torch.device(f"cuda:{device_idx}" if device_idx >= 0 else "cpu")


def device_ordinal(device: torch.device):
    return torch.zeros(1, device=device).get_device()


def same_storage(x, y):
    """Determine if tensors share the same storage or not"""
    x_ptrs = set(e.data_ptr() for e in x.view(-1))
    y_ptrs = set(e.data_ptr() for e in y.view(-1))
    return (x_ptrs <= y_ptrs) or (y_ptrs <= x_ptrs)


class MemReporter:
    """A utility class for reporting memory usage of PyTorch tensors by device.

    Features:
        - Tracks tensors in Python scope (excluding C++-managed buffers)
        - Reports memory in GB (base 1024) or GiB (base 1000) format
        - Identifies root tensors to avoid double-counting overlapping memory
        - Supports module-specific analysis or global tensor tracking

    Parameters
    ----------
    format_mode : Literal["GB", "GiB"]
        Memory unit format (GB=base1024, GiB=base1000)

    Attributes
    ----------
    tensors : dict
        Name-to-tensor mapping for all collected tensors
    device_map : defaultdict
        Maps devices to tensor names

    Examples
    --------
    >>> reporter = MemReporter(format_mode="GiB")
    >>> reporter.report()  # Analyze all tracked tensors
    >>> reporter.report(module=my_model)  # Analyze tensors in a specific module

    Notes
    -----
    - Does not track:
        * Tensors allocated in C++ (e.g. backward pass buffers)
        * Gradient tensors (.grad attributes)
    - Non-contiguous tensors may have inefficient pointer calculations
    """

    def __init__(self, format_mode: Literal["GB", "GiB"] = "GiB"):
        self.format_mode = format_mode
        self.tensors = {}
        self.device_map = defaultdict(list)

    @staticmethod
    def _sizeof(tensor):
        return tensor.element_size() * tensor.nelement()

    @staticmethod
    def _format_size(size_B, mode: Literal["GB", "GiB"] = "GB"):
        if mode == "GB":
            base = 1000
            prefix = ["K", "M", "G"]
        elif mode == "GiB":
            base = 1024
            prefix = ["Ki", "Mi", "Gi"]
        if size_B < base:
            return f"{size_B}", "B"
        elif size_B < base**2:
            return f"{size_B / base:.2f}", f"{prefix[0]}B"
        elif size_B < base**3:
            return f"{size_B / (base**2):.2f}", f"{prefix[1]}B"
        else:
            return f"{size_B / (base**3):.2f}", f"{prefix[2]}B"

    def _collect_tensors(self, module: Optional[nn.Module] = None):
        """Collect all tensor objects tracked by python

        NOTICE:
            - the buffers for backward which is implemented in C++ are
            not tracked by python's reference counting.
            - the gradients(.grad) of Parameters is not collected, and
            I don't know why.
        """
        gc.collect()
        if module is None:
            objects = gc.get_objects()
            n = len(self.tensors)  # Track number of tensors
            for obj in objects:
                if isinstance(obj, Tensor):
                    name = f"Tensor{n}"
                    self.tensors[name] = obj
                    self.device_map[obj.device].append(name)
                    n += 1
        else:
            for name, t in module.named_parameters():
                self.tensors[name] = t
                self.device_map[t.device].append(name)
        # print(f"{len(self.tensors)} tensor(s) collected")

    def _get_root_tensors(self, device, names):
        ptrs = {}
        for name, t in self.tensors.items():
            if device == t.device and name in names:
                # Get range of start and end pointers
                if not t.is_contiguous():
                    warn(f"Non-contiguous tensor {name} cannot be indexed efficiently")
                # (start, end)
                # ptrs[name] = (t.view(-1)[0].data_ptr(), t.view(-1)[-1].data_ptr())
                ptrs[name] = (
                    t.data_ptr(),
                    t.data_ptr() + t.nelement() * t.element_size() - 1,
                )
        # print(f"Collected {len(self.ptrs)} tensors")
        # Create a dependency graph of inclusion
        roots = []
        for name, this_ptrs in ptrs.items():
            counted = False
            new_roots = []
            for root in roots:
                root_ptrs = ptrs[root]
                if this_ptrs[0] >= root_ptrs[0] and this_ptrs[1] <= root_ptrs[1]:
                    # root supercedes this
                    new_roots.append(root)
                    counted = True
                elif this_ptrs[0] <= root_ptrs[0] and this_ptrs[1] >= root_ptrs[1]:
                    # this supercedes root
                    new_roots.append(name)
                    counted = True
                else:
                    # Other options that we didn't deal with...
                    # Independent
                    new_roots.append(root)
                    pass

            if not counted:
                new_roots.append(name)
            else:
                pass
            roots = new_roots
        return roots

    def report(self, module: Optional[nn.Module] = None):
        self._collect_tensors(module)
        for dev, names in self.device_map.items():
            total_size = 0
            roots = self._get_root_tensors(dev, names)
            print(f"Device {dev}")
            print("=" * 20)
            for name in names:
                if name in roots:
                    tensor = self.tensors[name]
                    shape = tuple(tensor.shape)
                    tsize = self._sizeof(tensor)
                    size, NB = self._format_size(tsize, self.format_mode)
                    print(f"{name}\t{shape}\t\t{size} {NB}")
                    total_size += tsize
            size, NB = self._format_size(total_size, self.format_mode)
            print(f"Total: {size} {NB}")



from warnings import warn
from copy import copy, deepcopy
from typing import Optional
from collections.abc import Mapping

import torch
from einops import rearrange, reduce, repeat

from .identity import Identity
from .namedlinop import NamedLinop
from .nameddim import ND, NS, NamedShape, Shape

__all__ = [
    "Rearrange",
    "SumReduce",
    "Repeat",
]


class Rearrange(NamedLinop):
    """Moves around dimensions."""

    def __init__(
        self,
        ipattern,
        opattern,
        ishape: Shape,
        oshape: Shape,
        axes_lengths: Optional[Mapping] = None,
    ):
        # assert len(ishape) == len(
        #     oshape
        # ), "Rearrange currently only supports pure dimension permutations"
        super().__init__(NS(ishape, oshape))
        self.ipattern = ipattern
        self.opattern = opattern
        axes_lengths = axes_lengths if axes_lengths is not None else {}
        self._shape.add("axes_lengths", axes_lengths)

    @property
    def axes_lengths(self):
        return self._shape.axes_lengths

    def forward(self, x):
        return self.fn(self, x)

    @staticmethod
    def fn(linop, x, /):
        axes_lengths = {str(k): v for k, v in linop.axes_lengths.items()}
        return rearrange(x, f"{linop.ipattern} -> {linop.opattern}", **axes_lengths)

    @staticmethod
    def adj_fn(linop, x, /):
        axes_lengths = {str(k): v for k, v in linop.axes_lengths.items()}
        return rearrange(x, f"{linop.opattern} -> {linop.ipattern}", **axes_lengths)

    def split_forward(self, ibatch, obatch):
        """TODO: Add compound shapes so splitting through rearrange can work."""
        warn(
            f"Splitting Rearrange linop with shape {self._shape} - splitting a rearrange may behave unusually."
        )
        new_axes_lengths = deepcopy(self.axes_lengths)
        for dim, slc in zip(self.ishape, ibatch):
            if dim in self.axes_lengths:
                n = self.axes_lengths[dim]
                new_axes_lengths[dim] = slicelen(n, slc)
        for dim, slc in zip(self.oshape, obatch):
            if dim in self.axes_lengths:
                n = self.axes_lengths[dim]
                new_axes_lengths[dim] = slicelen(n, slc)

        out = type(self)(
            self.ipattern, self.opattern, self.ishape, self.oshape, new_axes_lengths
        )
        return out

    def split_forward_fn(self, ibatch, obatch, /):
        """Rearranging is transparent to splitting"""
        return None

    def size(self, dim: str):
        """Rearranging does not determine any dimensions"""
        return None

    def size_fn(self, dim: str, /):
        """Rearranging does not determine any dimensions"""
        return None

    def normal(self, inner=None):
        if inner is None:
            return Identity(self.ishape)
        return super().normal(inner)


class SumReduce(NamedLinop):
    """Wrapper for einops' reduce,

    Adjoint of Repeat
    """

    def __init__(self, ishape, oshape):
        """
        ipattern : string
            Input shape spec, einops style
        opattern : string
            Output shape spec, einops style
        """
        super().__init__(NS(ishape, oshape))
        assert len(self.oshape) < len(self.ishape), (
            f"Reduce must be over at least one dimension: got {self.ishape} -> {self.oshape}"
        )

    def forward(self, x):
        return self.fn(self, x)

    @staticmethod
    def fn(linop, x, /):
        x = reduce(x, f"{linop.ipattern} -> {linop.opattern}", "sum")
        return x

    @staticmethod
    def adj_fn(linop, x, /):
        x = repeat(x, f"{linop.opattern} -> {linop.adj_ipattern}")
        return x

    def split_forward(self, ibatch, obatch):
        return self

    def split_forward_fn(self, ibatch, obatch, /):
        """Reducing is transparent to splitting"""
        return None

    def size(self, dim: str):
        """Reducing does not determine any dimensions"""
        return None

    def size_fn(self, dim: str, /, ipattern, opattern, size_spec):
        """Reducing does not determine any dimensions"""
        return None

    def adjoint(self):
        broadcast_dims = [d for d in self.ishape if d not in self.oshape]
        n_repeats = {d: 1 for d in broadcast_dims}
        return Repeat(n_repeats, self._shape.H, None, broadcast_dims)

    def normal(self, inner=None):
        pre = copy(self)
        post = self.adjoint()
        # New post output shape (post = Repeat)
        # If dimension is not summed over (i.e. it is in pre_adj_ishape) , it stays the same
        # Otherwise, if dimension is summed over, its name changes
        # This automatically updates the axes_lengths as well.
        post.oshape = tuple(
            d.next_unused(self.ishape) if d not in post.ishape else d
            for d in post.oshape
        )
        shape_updates = {
            d: new_d for d, new_d in zip(pre.ishape, post.oshape) if d != new_d
        }
        if inner is not None:
            pre.oshape = inner.ishape
            post.ishape = inner.oshape
            inner_shape_updates = getattr(inner, "_shape_updates", {})
            shape_updates.update(inner_shape_updates)
            normal = post @ inner @ pre
            normal._shape_updates = shape_updates
        else:
            normal = post @ pre
            normal._shape_updates = shape_updates
        return normal

    @property
    def adj_ishape(self):
        return self.fill_singleton_dims(self.ishape, self.oshape)

    @property
    def adj_ipattern(self):
        return " ".join(str(d) if d is not None else "()" for d in self.adj_ishape)

    @property
    def ipattern(self):
        return " ".join(str(d) for d in self.ishape)

    @property
    def opattern(self):
        return " ".join(str(d) for d in self.oshape)

    @staticmethod
    def fill_singleton_dims(ishape, oshape):
        out = []
        for idim in ishape:
            if idim in oshape:
                out.append(idim)
            else:
                out.append(None)
        return tuple(out)


class Repeat(NamedLinop):
    """Unsqueezes and expands a tensor along dim"""

    def __init__(
        self, n_repeats: Mapping, ishape, oshape, broadcast_dims: Optional[list] = None
    ):
        super().__init__(NS(ishape, oshape))
        assert len(self.oshape) > len(self.ishape), (
            f"Repeat must add at least one dimension: got {self.ishape} -> {self.oshape}"
        )
        self._shape.add("axes_lengths", n_repeats)
        # self.axes_lengths = n_repeats
        # self.axes_lengths = {ND.infer(k): v for k, v in self.axes_lengths.items()}
        broadcast_dims = broadcast_dims if broadcast_dims is not None else []
        self._shape.add("broadcast_dims", broadcast_dims)

    @property
    def axes_lengths(self):
        return self._shape.axes_lengths

    @property
    def broadcast_dims(self):
        return self._shape.broadcast_dims

    def forward(self, x):
        return self.fn(self, x)

    @staticmethod
    def fn(linop, x, /):
        x = repeat(
            x,
            f"{linop.ipattern} -> {linop.opattern}",
            **{str(k): v for k, v in linop.axes_lengths.items()},
        )
        return x

    @staticmethod
    def adj_fn(linop, x, /):
        x = reduce(x, f"{linop.opattern} -> {linop.ipattern}", "sum")
        return x

    def split_forward(self, ibatch, obatch):
        """Repeat fewer times, depending on the size of obatch"""
        new_axes_lengths = deepcopy(self.axes_lengths)
        for dim, slc in zip(self.oshape, obatch):
            if dim in self.axes_lengths and dim not in self.broadcast_dims:
                self.axes_lengths[dim] = slicelen(self.size(dim), slc)
        return type(self)(
            new_axes_lengths, self.ishape, self.oshape, self.broadcast_dims
        )

    def split_forward_fn(self, ibatch, obatch, /):
        """No data to split"""
        return None

    def size(self, dim: str):
        return self.size_fn(dim)

    def size_fn(self, dim, /):
        if dim in self.broadcast_dims:
            return None
        return self.axes_lengths.get(dim, None)

    def adjoint(self):
        return SumReduce(self._shape.H, None)

    def normal(self, inner=None):
        pre = copy(self)
        post = self.adjoint()
        if inner is not None:
            pre.oshape = inner.ishape
            post.ishape = inner.oshape
            return post @ inner @ pre
        # No updated dims because Repeat -> SumReduce
        # Gets rid of the new dimensions immediately
        # TODO: simplify this more?
        return post @ pre

    @property
    def adj_ishape(self):
        return self.fill_singleton_dims(self.oshape, self.ishape)

    @property
    def adj_ipattern(self):
        return " ".join(str(d) if d is not None else "()" for d in self.adj_ishape)

    @property
    def ipattern(self):
        return " ".join(str(d) for d in self.ishape)

    @property
    def opattern(self):
        return " ".join(str(d) for d in self.oshape)

    @staticmethod
    def fill_singleton_dims(ishape, oshape):
        out = []
        for idim in ishape:
            if idim in oshape:
                out.append(idim)
            else:
                out.append(None)
        return tuple(out)


def slicelen(n, slc):
    return len(range(*slc.indices(n)))



import pytest

import torch

import torchlinops.functional as F
from torchlinops import Sampling
from torchlinops.tests.test_base import BaseNamedLinopTests


class TestSampling(BaseNamedLinopTests):
    equality_check = "approx"

    isclose_kwargs = dict(rtol=1e-4)

    @pytest.fixture(scope="class")
    def linop_input_output(self):
        N = 64
        ndim = 2
        R, K = 13, 17
        B = 3  # Batch
        idx = torch.randint(0, N - 1, (R, K, ndim))
        x = torch.randn(B, N, N)
        y = torch.randn(B, R, K)
        linop = Sampling.from_stacked_idx(idx, (N, N), ("R", "K"))
        return linop, x, y


def test_sampling_slc():
    N = 64
    ndim = 2
    R, K = 13, 17
    B = 3  # Batch
    idx = torch.randint(0, N - 1, (R, K, ndim))
    x = torch.randn(B, N, N)

    linop = Sampling.from_stacked_idx(idx, (N, N), ("R", "K"))

    linop_split = linop.split(linop, {"R": slice(2, 5), "K": slice(4, 10)})
    Ax = linop(x)
    Ax_split = linop_split(x)
    assert (Ax[:, 2:5, 4:10] == Ax_split).all()



from collections.abc import Callable
from typing import Union, Optional, Tuple
from typing_extensions import Self
from torch import Tensor

import traceback
from pprint import pformat

import torch
import torch.nn as nn
from tqdm import tqdm

from .namedlinop import NamedLinop
from .identity import ShapeSpec
from .nameddim import ND, NS, Shape
from .split import split_linop

from torchlinops.utils import batch_iterator, dict_product, INDENT

__all__ = ["Batch"]


class Batch(NamedLinop):
    def __init__(
        self,
        linop: NamedLinop,
        input_device: torch.device,
        output_device: torch.device,
        input_dtype: Union[str, torch.dtype],
        output_dtype: Union[str, torch.dtype],
        input_shape: Optional[Shape] = None,
        output_shape: Optional[Shape] = None,
        pbar: bool = False,
        name: Optional[str] = None,
        post_batch_hook: Optional[Callable] = None,
        **batch_sizes,
    ):
        """
        hook : Callable, optional
            Function that takes in the newly-created batch object and does stuff
        """
        # TODO: Should batch even have a shape???
        super().__init__(NS(linop.ishape, linop.oshape), name=name)

        self.linop = linop
        if input_shape is not None:
            self.linop = self.linop @ ShapeSpec(input_shape)
        if output_shape is not None:
            self.linop = ShapeSpec(output_shape) @ self.linop
        self.input_device = input_device
        self.output_device = output_device
        self.input_dtype = input_dtype
        self.output_dtype = output_dtype
        self.pbar = pbar
        # self.name = name if name is not None else ""
        self.batch_sizes = batch_sizes
        self.post_batch_hook = post_batch_hook
        self.sizes = {dim: self.linop.size(dim) for dim in self.linop.dims}
        self.setup_batching()

    def setup_batching(self, hook: Optional[Callable] = None):
        _linops, _input_batches, _output_batches = split_linop(
            self.linop,
            self.batch_sizes,
        )
        self._linops = nn.ModuleList(_linops.flatten().tolist())
        self._input_batches = _input_batches.flatten().tolist()
        self._output_batches = _output_batches.flatten().tolist()
        self._shape = NS(self.linop.ishape, self.linop.oshape)
        super().reset_adjoint_and_normal()
        if self.post_batch_hook is not None:
            self.post_batch_hook(self)

    def to(self, device: torch.device | str) -> Self:
        self.input_device = device
        self.output_device = device
        # self.linop.to(device)
        # Avoid copying the _linops by doing super() first
        super().to(device)
        self.setup_batching()
        return self

    def forward(self, x: torch.Tensor):
        # Complete the size specifications
        for dim, total in zip(self.ishape, x.shape):
            self.sizes[dim] = total

        y = torch.zeros(
            tuple(self.sizes[dim] for dim in self.oshape),
            dtype=self.output_dtype,
            device=self.output_device,
        )
        for linop, in_batch, out_batch in tqdm(
            zip(self._linops, self._input_batches, self._output_batches),
            total=len(self._linops),
            desc=f"Batch({self.name}: {self.batch_sizes})",
            disable=(not self.pbar),
        ):
            try:
                xbatch = x[in_batch]
                ybatch = linop(xbatch)
                y[out_batch] += ybatch
            except RuntimeError:
                print(
                    f"linop: {linop}, in_batch: {in_batch}, out_batch: {out_batch}, self.batch_sizes: {self.batch_sizes}"
                )
                raise
        return y

    @property
    def H(self):
        if self._adjoint is None:
            try:
                _adjoint = self.adjoint()
                _adjoint._adjoint = [self]
                self._adjoint = [_adjoint]
            except AttributeError as e:
                traceback.print_exc()
                raise e
        return self._adjoint[0]

    def adjoint(self):
        batch_sizes = {str(k): v for k, v in self.batch_sizes.items()}
        adj = type(self)(
            linop=self.linop.H,
            input_device=self.output_device,
            output_device=self.input_device,
            input_dtype=self.output_dtype,
            output_dtype=self.input_dtype,
            name=self.name + ".H",
            pbar=self.pbar,
            post_batch_hook=self.post_batch_hook,
            **batch_sizes,
        )
        return adj

    @property
    def N(self):
        if self._normal is None:
            try:
                _normal = self.normal()
                self._normal = [_normal]
            except AttributeError as e:
                traceback.print_exc()
                raise e
        return self._normal[0]

    def normal(self, inner=None):
        normal_linop = self.linop.N
        # Collect shape updates from computing the normal
        shape_updates = getattr(normal_linop, "_shape_updates", {})
        new_batch_sizes = self.batch_sizes.copy()
        for d, nd in shape_updates.items():
            if d in self.batch_sizes:
                new_batch_sizes[shape_updates[d]] = self.batch_sizes[d]
            elif str(d) in self.batch_sizes:
                new_batch_sizes[str(shape_updates[d])] = self.batch_sizes[str(d)]
        batch_size_kwargs = {str(k): v for k, v in new_batch_sizes.items()}
        normal = type(self)(
            linop=self.linop.N,
            input_device=self.input_device,
            output_device=self.input_device,
            input_dtype=self.input_dtype,
            output_dtype=self.input_dtype,
            name=self.name + ".N",
            pbar=self.pbar,
            post_batch_hook=self.post_batch_hook,
            **batch_size_kwargs,
        )

        return normal

    @staticmethod
    def fn(self, x, /, data):
        """TODO: Functional interface
        Specify data as a tuple of data entries, one for each linop in linops"""
        raise NotImplementedError(f"Batched functional interface not available yet.")
        sizes = {}
        for dim in self.linop.dims:
            sizes[dim] = self.linop.size_fn(dim, data)
        for dim, total in zip(self.ishape, x.shape):
            sizes[dim] = total
        batch_iterators = self._make_batch_iterators(sizes, self.batch_sizes)
        ishapes = [linop.ishape for linop in self.linop.flatten()]
        oshapes = [linop.oshape for linop in self.linop.flatten()]

        y = torch.zeros(
            tuple(sizes[dim] for dim in self.oshape),
            dtype=self.output_dtype,
            device=self.output_device,
        )
        for tile in tqdm(
            dict_product(batch_iterators),
            desc=f"Batch({self.batch_sizes})",
            disable=(not self.pbar),
        ):
            ibatches = [
                [tile.get(dim, slice(None)) for dim in ishape] for ishape in ishapes
            ]
            obatches = [
                [tile.get(dim, slice(None)) for dim in oshape] for oshape in oshapes
            ]
            split_data = self.linop.split_fn(*ibatches, *obatches, *data)
            xbatch = x[ibatches[0]].to(self.input_device)
            ybatch = self.linop.fn(xbatch, split_data)
            y[obatches[-0]] += ybatch
        return y

    @staticmethod
    def adj_fn(self, x, /, data):
        raise NotImplementedError("Batched linop has no adjoint (yet).")

    def size(self, dim):
        return self.linop.size(dim)

    def size_fn(self, dim, /, data=None):
        raise NotImplementedError()

    def __repr__(self):
        """Helps prevent recursion error caused by .H and .N"""
        output = ""
        output += INDENT.indent(self.name + self._suffix) + "(\n"
        with INDENT:
            output += repr(self.linop)
            output += ", "
            output += repr(pformat(self.batch_sizes)).strip("'")
            output += "\n"
        output += INDENT.indent(")")
        return output



from collections.abc import Iterable
from torch import Tensor
import torch

__all__ = ["device_ordinal", "same_storage", "idx2dev"]


def device_ordinal(device: torch.device) -> int:
    """Get the integer representing the device ordinal.

    Parameters
    ----------
    device : torch.device
        The device in question.

    Returns
    -------
    int
        The device ordinal - 0+ if GPU, -1 if cpu
    """
    return torch.zeros(1, device=device).get_device()


def same_storage(x: Tensor, y: Tensor):
    """Determine if tensors share the same storage or not"""
    x_ptrs = set(e.data_ptr() for e in x.view(-1))
    y_ptrs = set(e.data_ptr() for e in y.view(-1))
    return (x_ptrs <= y_ptrs) or (y_ptrs <= x_ptrs)


def idx2dev(idx: int | list):
    if isinstance(idx, Iterable):
        return [idx2dev(i) for i in idx]
    return torch.device(f"cuda:{idx}" if int(idx) >= 0 else "cpu")



import pytest

import torch
from torchlinops import Dense, Batch


def test_batch_normal_adjoint():
    ishape = ("B", "N")
    oshape = ("B", "M")
    B = 10
    M, N = (3, 7)
    weight = torch.randn(B, M, N)
    weightshape = ("B", "M", "N")
    device = "cpu"
    A_nobatch = Dense(weight, weightshape, ishape, oshape)
    A = Batch(A_nobatch, device, device, weight.dtype, weight.dtype, **{"N": 1})
    x = torch.randn(B, N)
    y = torch.randn(B, M)

    # Test Forward
    Ax = A(x)
    Ax_ref = A_nobatch(x)
    assert Ax_ref.allclose(Ax)

    # Test Normal
    AN = A.N
    ANx = AN(x)
    ANx_ref = A_nobatch.N(x)
    assert len(AN._linops) == N * N
    assert ANx_ref.allclose(ANx)

    # Test Adjoint
    AH = A.H
    AHy = AH(y)
    AHy_ref = A_nobatch.H(y)
    assert len(AH._linops) == N
    assert AHy_ref.allclose(AHy)


def test_batch_chain_normal_adjoint(): ...



import pytest

import torch
import sigpy as sp

from torchlinops.functional import ungrid, grid
from torchlinops.utils import to_pytorch, from_pytorch

PYTEST_GPU_MARKS = [
    pytest.mark.gpu,
    pytest.mark.skipif(
        not torch.cuda.is_available(), reason="GPU is required but not available"
    ),
]


@pytest.mark.parametrize("dev", ["cpu", pytest.param("cuda", marks=PYTEST_GPU_MARKS)])
@pytest.mark.parametrize("dtype", ["real", "complex"])
def test_exact_int_loc_ungrid(dev, dtype):
    device = torch.device(dev)
    dtype = torch.complex64 if dtype == "complex" else torch.float32

    vals = torch.arange(1, 9).to(device).to(dtype)
    width = 4.0
    grid_size = (len(vals),)
    locs = torch.tensor([[4.0]], device=device)
    interp = ungrid(vals, locs, width=width, kernel="kaiser_bessel")

    # Test against sigpy
    vals = from_pytorch(vals)
    # locs = from_pytorch(locs.flip(dims=(-1,)).contiguous())
    locs = from_pytorch(locs)

    # Sigpy quirks:
    # - Circular padding
    # - weight is product of 1D kernel evals along each coordinate axis
    interp_sp = sp.interpolate(vals, locs, kernel="kaiser_bessel", width=width)
    interp_sp = to_pytorch(interp_sp)

    assert interp.allclose(interp_sp)


@pytest.mark.parametrize("dev", ["cpu", pytest.param("cuda", marks=PYTEST_GPU_MARKS)])
@pytest.mark.parametrize("dtype", ["real", "complex"])
def test_exact_int_loc_grid(dev, dtype):
    device = torch.device(dev)
    dtype = torch.complex64 if dtype == "complex" else torch.float32

    vals = torch.tensor([[1.0]], device=device)
    width = 4.0
    grid_size = (8,)
    locs = torch.tensor([[4.0]], device=device)
    interp = grid(vals, locs, grid_size, width=width, kernel="kaiser_bessel")

    # Test against sigpy
    vals = from_pytorch(vals)
    # locs = from_pytorch(locs.flip(dims=(-1,)).contiguous())
    locs = from_pytorch(locs)

    # Sigpy quirks:
    # - Circular padding
    # - weight is product of 1D kernel evals along each coordinate axis
    interp_sp = sp.gridding(vals, locs, grid_size, kernel="kaiser_bessel", width=width)
    interp_sp = to_pytorch(interp_sp)

    assert interp.allclose(interp_sp)



import pytest

import torch

from torchlinops import Stack, Dense, Diagonal, ND, Dim
from einops import einsum

from torchlinops.tests.test_base import BaseNamedLinopTests
from torchlinops.utils import inner


def test_stack_init():
    P, Q = 3, 4
    ishape = Dim("Q")
    oshape = Dim("P")

    A = Dense(torch.randn(P, Q), ("P", "Q"), ishape, oshape)
    B = Dense(torch.randn(P, Q), ("P", "Q"), ishape, oshape)
    C = Stack(A, B, idim_and_idx=("N", 0))
    C2 = Stack(A, B, odim_and_idx=("N", 1))
    C3 = Stack(A, B, idim_and_idx=("N", 0), odim_and_idx=("N", 1))


@pytest.fixture
def denselinops():
    P, Q = 3, 4
    ishape = Dim("Q")
    oshape = Dim("P")
    wA = torch.randn(P, Q)
    wB = torch.randn(P, Q)
    A = Dense(wA, ("P", "Q"), ishape, oshape)
    B = Dense(wB, ("P", "Q"), ishape, oshape)
    return A, B


def test_stack_horizontal(denselinops):
    A, B = denselinops
    idim = 0
    # Horizontal stack
    C = Stack(A, B, idim_and_idx=("N", idim))
    assert C.size("N") == len([A, B])
    x = torch.randn(C.size("N"), C.size("Q"))
    Cx = C(x)
    xs = x.tensor_split(len([A, B]), idim)
    xs = [xi.squeeze(idim) for xi in xs]
    Cx_ref = A(xs[0]) + B(xs[1])
    assert Cx.allclose(Cx_ref)


def test_stack_vertical(denselinops):
    A, B = denselinops
    odim = 1
    # Vertical stack
    C = Stack(A, B, odim_and_idx=("N", odim))
    assert C.size("N") == len([A, B])
    x = torch.randn(C.size("Q"))
    Cx = C(x)
    Cx_ref = torch.stack((A(x), B(x)), dim=odim)
    assert Cx.allclose(Cx_ref)


def test_stack_diagonal(denselinops):
    A, B = denselinops
    idim = 0
    odim = 1
    # Diagonal stack
    C = Stack(A, B, idim_and_idx=("N", idim), odim_and_idx=("M", odim))
    assert C.size("N") == len([A, B])
    assert C.size("M") == len([A, B])
    x = torch.randn(C.size("N"), C.size("Q"))
    Cx = C(x)
    xs = x.tensor_split(len([A, B]), idim)
    xs = [xi.squeeze(idim) for xi in xs]
    Cx_ref = torch.stack((A(xs[0]), B(xs[1])), dim=odim)
    assert Cx.allclose(Cx_ref)


def test_stack_split(denselinops):
    A, B = denselinops
    idim = 0
    odim = 1
    # Vertical stack
    C = Stack(A, B, odim_and_idx=("M", odim))

    # Slice along non-stack dim
    C1 = C.split(C, {"P": slice(0, 1)})
    assert (C1[0].weight == A.weight[:1]).all()

    # Slice along stack dim
    C2 = C.split(C, {"M": slice(0, 2)})
    assert isinstance(C2, Stack)
    assert (C2[0].weight == A.weight).all()
    assert (C2[1].weight == B.weight).all()

    # Horizontal stack
    C = Stack(A, B, idim_and_idx=("N", idim))
    # Slice along non-stack dim
    C3 = C.split(C, {"Q": slice(1, 2)})
    assert (C3[0].weight == A.weight[:, 1:2]).all()
    assert (C3[1].weight == B.weight[:, 1:2]).all()

    # Slice along stack dim
    C4 = C.split(C, {"N": slice(1, 2)})
    assert isinstance(C4, Stack)
    assert (C4[0].weight == B.weight).all()


def test_stack_adjoint(denselinops):
    A, B = denselinops
    idim = 0
    odim = 1
    rtol = 1e-3
    # Vertical stack
    C = Stack(A, B, odim_and_idx=("M", odim))
    x = torch.randn(*(C.size(d) for d in C.ishape))
    y = torch.randn(*(C.size(d) for d in C.oshape))
    assert torch.allclose(inner(y, C(x)), inner(C.H(y), x), rtol=rtol)

    # Horizontal stack
    C = Stack(A, B, idim_and_idx=("N", idim))
    x = torch.randn(*(C.size(d) for d in C.ishape))
    y = torch.randn(*(C.size(d) for d in C.oshape))
    assert torch.allclose(inner(y, C(x)), inner(C.H(y), x), rtol=rtol)

    # Diagonal stack
    C = Stack(A, B, idim_and_idx=("N", idim), odim_and_idx=("M", odim))
    CH = C.H
    x = torch.randn(*(C.size(d) for d in C.ishape))
    y = torch.randn(*(C.size(d) for d in C.oshape))
    assert torch.allclose(inner(y, C(x)), inner(C.H(y), x), rtol=rtol)


def test_stack_normal(denselinops):
    # TODO: Add some kind of assertions here?
    A, B = denselinops
    idim = 0
    odim = 1
    # Vertical stack
    C = Stack(A, B, odim_and_idx=("M", odim))
    x = torch.randn(*(C.size(d) for d in C.ishape))
    assert torch.allclose(C.N(x), A.N(x) + B.N(x))

    # Horizontal stack
    C = Stack(A, B, idim_and_idx=("N", idim))
    x = torch.randn(*(C.size(d) for d in C.ishape))
    CNx = C.N(x)
    x0, x1 = x[0], x[1]
    ANx = A.N(x0)
    AHBx = A.H(B(x1))
    BHAx = B.H(A(x0))
    BNx = B.N(x1)
    y0 = ANx + AHBx
    y1 = BHAx + BNx
    out = torch.stack([y0, y1], dim=idim)
    assert torch.allclose(CNx, out)

    # Diagonal stack
    C = Stack(A, B, idim_and_idx=("N", idim), odim_and_idx=("M", odim))
    x = torch.randn(*(C.size(d) for d in C.ishape))
    CNx = C.N(x)
    x0, x1 = x[0], x[1]
    y0 = A.N(x0)
    y1 = B.N(x1)
    out = torch.stack([y0, y1], dim=odim)
    assert torch.allclose(CNx, out)



#!/usr/bin/env python
#
import glob
import logging
import json


logger = logging.getLogger("torchmri")


def main():
    trace_files = glob.glob("./worker_*_trace.json")
    logger.info(f"Found trace files ", trace_files)
    merge_traces(trace_files, "merged_trace.json")


def merge_traces(input_files, output_file):
    """

    Example usage
    trace_files = glob.glob("./worker_*_trace.json")
    merge_traces(trace_files, "merged_trace.json")
    """
    all_traces = {}
    all_traces["traceEvents"] = []

    for i, trace_file in enumerate(input_files):
        with open(trace_file, "r") as f:
            trace_data = json.load(f)
            events = trace_data["traceEvents"]
            for event in events:
                event["pid"] = str(event["pid"]) + f"_{i}"

            # Important step
            all_traces["traceEvents"].extend(events)
            # Less important steps
            all_traces["schemaVersion"] = trace_data["schemaVersion"]
            all_traces["deviceProperties"] = trace_data["deviceProperties"]
    all_traces["traceName"] = str(output_file)

    # Write combined traces to a new file
    with open(output_file, "w") as f:
        json.dump(all_traces, f)


if __name__ == "__main__":
    main()



import copy

from torchlinops import NamedDimension, ND


def test_nameddim():
    C = NamedDimension("C")
    C1 = copy.copy(C)
    C1 = C + 1
    assert C1.i == 1

    C3 = NamedDimension.infer(C)
    assert C3.name == "C"


def test_nameddim_repr():
    A = NamedDimension("A")
    assert repr(A) == "A"
    B = NamedDimension("B", i=1)
    assert repr(B) == "B1"


def test_nameddim_infer():
    Nx = NamedDimension.infer("Nx")
    assert Nx.i == 0 and Nx.name == "Nx"
    P1 = NamedDimension.infer("P1")
    assert P1.i == 1 and P1.name == "P"
    ishape = ("C", "Nx", "Ny")
    ishape2 = NamedDimension.infer(ishape)
    assert ishape2[1] == NamedDimension("Nx")


def test_nameddim_add():
    K = NamedDimension("K")
    K4 = K + 4
    assert K4.i == 4, K4.name == "K"


def test_nameddim_ellipsis():
    ellipsis = ND("...")
    assert ellipsis == "..."
    assert ellipsis == ND("...")


def test_nameddim_hash_lookup():
    foo = {"P": 1}
    assert foo[ND("P")] == 1

    bar = {ND("Q"): 2}
    assert bar["Q"] == 2



import logging

from textwrap import indent

__all__ = ["setup_console_logger", "Indenter", "INDENT"]


def setup_console_logger(
    logger,
    log_level: int,
    fmt: str = "%(asctime)s | %(name)s | %(levelname)s | %(message)s",
    handler_cls=logging.StreamHandler,
):
    logger.setLevel(log_level)
    formatter = logging.Formatter(fmt)
    for handler in logger.handlers:
        if isinstance(handler, handler_cls):
            handler.setFormatter(formatter)
            return logger
    handler = handler_cls()
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    return logger


class Indenter:
    """Special context manager for indented formatting

    Example usage:

    with Indenter() as ident:
        ident.print("[DRY RUN]") # Not indented
        with ident:
            ident.print(pipeline.pprint()) # Indented one level
    """

    def __init__(self, start_level: int = -1):
        self.level = start_level

    def indent(self, text):
        for _ in range(self.level):
            text = indent(text, "    ")
        return text

    def print(self, text):
        print(self.indent(text))

    def __enter__(self):
        self.level += 1
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.level -= 1


# Singleton for global indentation control
INDENT = Indenter(start_level=0)



from typing import Optional

__all__ = ["NDList"]


class NDList:
    """DEPRECATED in favor of np.ndarray(..., dtype=object)"""

    def __init__(self, shape, default_value=None, labels: Optional[tuple] = None):
        """
        Initializes an n-dimensional nested with the given shape.
        :param shape: Tuple representing the shape of the array.
        :param default_value: The default value to fill the array.
        """
        self.shape = shape
        self.data = self._create_nd_array(shape, default_value)
        if labels is None:
            self.labels = (None,) * len(self.shape)
        else:
            if len(labels) != len(self.shape):
                raise ValueError(
                    f"If specified, labels must have same length as shape but got {len(labels)} != {len(shape)}"
                )
            self.labels = labels

    def _create_nd_array(self, shape, default_value, depth=0):
        if depth == len(shape) - 1:
            return [default_value] * shape[depth]
        return [
            self._create_nd_array(shape, default_value, depth + 1)
            for _ in range(shape[depth])
        ]

    def _get_nested(self, indices):
        if not isinstance(indices, tuple):
            indices = (indices,)
        elem = self.data
        for index in indices:
            if index == slice(None):
                return [
                    self._get_nested((i,) + indices[1:]) for i in range(self.shape[0])
                ]
            elem = elem[index]
        return elem

    def _set_nested(self, indices, value):
        if not isinstance(indices, tuple):
            indices = (indices,)
        if indices[0] == slice(None):
            for i in range(self.shape[0]):
                self._set_nested((i,) + indices[1:], value)
            return
        elem = self.data
        for index in indices[:-1]:
            elem = elem[index]
        elem[indices[-1]] = value

    def __getitem__(self, index):
        return self._get_nested(index)

    def __setitem__(self, index, value):
        self._set_nested(index, value)

    def __repr__(self):
        return repr(self.data)



from types import SimpleNamespace
from torch import Tensor
from torchlinops.linops.pad_last import PadLast, pad_to_size, crop_slice_from_pad


__all__ = ["center_pad", "center_crop"]


def center_pad(x: Tensor, im_size: tuple[int, ...], pad_im_size: tuple[int, ...]):
    """
    Center pads the input tensor to the specified size.

    Parameters
    ----------
    x : Tensor
        The input tensor to be padded.
    im_size : tuple of int
        The original size of the input tensor.
    pad_im_size : tuple of int
        The desired size after padding.

    Returns
    -------
    Tensor
        The padded tensor.

    Notes
    -----
    This function uses the `PadLast` operator to perform the padding operation.
    It calculates the necessary padding dimensions and applies the padding
    to ensure that the output tensor matches the `pad_im_size` while preserving
    the center of the original tensor.
    """
    ndim = len(im_size)
    pad = pad_to_size(im_size, pad_im_size)
    crop_slice = crop_slice_from_pad(pad)
    pad_linop = SimpleNamespace(
        im_size=im_size,
        pad_im_size=pad_im_size,
        D=ndim,
        pad=pad,
        crop_slice=crop_slice,
    )
    return PadLast.fn(pad_linop, x)


def center_crop(x: Tensor, im_size: tuple[int, ...], crop_im_size: tuple[int, ...]):
    """
    Center crops the input tensor to the specified size.

    Parameters
    ----------
    x : Tensor
        The input tensor to be cropped.
    im_size : tuple of int
        The original size of the input tensor.
    crop_im_size : tuple of int
        The desired size after cropping.

    Returns
    -------
    Tensor
        The cropped tensor.

    Notes
    -----
    This function uses the `PadLast` operator to perform the cropping operation.
    It calculates the necessary cropping dimensions and applies the cropping
    from the center of the original tensor to achieve the specified `crop_im_size`.
    """
    ndim = len(im_size)
    pad = pad_to_size(crop_im_size, im_size)
    crop_slice = crop_slice_from_pad(pad)
    crop_linop = SimpleNamespace(
        im_size=crop_im_size,
        pad_im_size=im_size,
        D=ndim,
        pad=pad,
        crop_slice=crop_slice,
    )
    return PadLast.adj_fn(crop_linop, x)



import pytest

import torch
from torchlinops import ToDevice


@pytest.mark.gpu
@pytest.mark.skipif(
    not torch.cuda.is_available(), reason="GPU is required but not available"
)
def test_todevice():
    idevice = torch.device("cpu")
    odevice = torch.device("cuda:0")
    D2D = ToDevice(idevice, odevice)
    x = torch.randn(3, 4)
    y = D2D(x)
    assert y.device == odevice

    z = D2D.H(y)
    assert z.device == idevice

    w = D2D.N(x)
    assert w.device == x.device
    print(D2D)



import pytest

import torch

from torchlinops import SumReduce, Repeat

from torchlinops.utils import is_adjoint
from torchlinops.tests.test_base import BaseNamedLinopTests


class TestSumReduce(BaseNamedLinopTests):
    equality_check = "approx"

    isclose_kwargs = {"rtol": 1e-4}

    @pytest.fixture(scope="class", params=["fullshape", "ellipses"])
    def linop_input_output(self, request):
        x = torch.randn(5, 2, 3)
        y = torch.randn(5, 2)
        if request.param == "fullshape":
            linop = SumReduce(("A", "B", "C"), ("A", "B"))
        elif request.param == "ellipses":
            linop = SumReduce(("...", "C"), ("...",))
        return linop, x, y


def test_reduce_repeat():
    M, N = 5, 7
    x = torch.randn(M, N)
    y = torch.randn(N)

    A = SumReduce(("M", "N"), ("N",))
    B = Repeat({"M": M}, ("N",), ("M", "N"))

    assert is_adjoint(A, x, y)
    assert is_adjoint(B, y, x)



from ._adjoint_helpers import *
from ._apply_struct import *
from ._batching import *
from ._cupy_pytorch import *
from ._defaults import *
from ._device import *
from ._dispatch import *
from ._event import *
from ._fft import *
from ._log import *
from ._ndlist import *
from ._pad import *
from ._triton import *



"""Truncate/Pad
Maybe replace with more generic slicing linop later
"""

from copy import copy

from torchlinops.utils import end_pad_with_zeros
from .namedlinop import NamedLinop
from .nameddim import NS, Shape

__all__ = ["Truncate", "PadDim"]


class Truncate(NamedLinop):
    def __init__(
        self,
        dim: int,
        length: int,
        ishape: Shape,
        oshape: Shape,
    ):
        self.dim = dim
        self.length = length

        # Create the slices
        self.slc = [slice(None)] * len(ishape)
        self.slc[dim] = slice(0, self.length)
        self.slc = tuple(self.slc)

        self.end_slc = [slice(None)] * len(oshape)
        self.end_slc[dim] = slice(-self.length, None)
        self.end_slc = tuple(self.slc)
        super().__init__(NS(ishape, oshape))

    def forward(self, x):
        return self.fn(self, x)

    @staticmethod
    def fn(linop, x, /):
        return x[linop.slc]

    @staticmethod
    def adj_fn(linop, y, /):
        return end_pad_with_zeros(y, linop.dim, linop.length)

    @staticmethod
    def normal_fn(linop, x, /):
        x[linop.end_slc] = 0.0
        return x

    def split_forward(self, ibatch, obatch):
        if ibatch[self.dim] != slice(None) or obatch[self.dim] != slice(None):
            raise ValueError("Cannot slice a Truncate linop along truncation dimension")
        return type(self)(self.dim, self.length, self.ishape, self.oshape)

    def split_forward_fn(self, ibatch, obatch, /, data=None):
        if ibatch[self.dim] != slice(None) or obatch[self.dim] != slice(None):
            raise ValueError("Cannot slice a Truncate linop along truncation dimension")
        return None

    # Linop changes relative size, but can't determine the size itself
    def size(self, dim):
        return None

    def size_fn(self, dim, /, data=None):
        return None

    def adjoint(self):
        return PadDim(self.dim, self.length, self.oshape, self.ishape)

    def normal(self, inner=None):
        """Diagonal in all dims except the last one"""
        pre = copy(self)
        post = self.adjoint()
        if inner is None:
            return post @ pre
        pre.oshape = inner.ishape
        post.ishape = inner.oshape
        new_oshape = list(inner.oshape)
        new_oshape[self.dim] = post.oshape[self.dim]
        post.oshape = tuple(new_oshape)
        return post @ inner @ pre

    @staticmethod
    def is_in_slice(a_slice, idx):
        """TODO: unused"""
        if idx < a_slice.start or idx >= a_slice.stop:
            return False
        step = a_slice.step if a_slice.step else 1
        if (idx - a_slice.start) % step == 0:
            return True
        else:
            return False


class PadDim(NamedLinop):
    def __init__(self, dim, length, ishape, oshape):
        self.dim = dim
        self.length = length
        # Create the slices
        self.slc = [slice(None)] * len(ishape)
        self.slc[dim] = slice(0, self.length)
        self.slc = tuple(self.slc)

        self.end_slc = [slice(None)] * len(oshape)
        self.end_slc[dim] = slice(-self.length, 0)
        self.end_slc = tuple(self.end_slc)
        super().__init__(NS(ishape, oshape))

    def forward(self, x):
        return self.fn(self, x)

    def adjoint(self):
        return Truncate(self.dim, self.length, self.oshape, self.ishape)

    def normal(self, inner=None):
        """Diagonal in all dims except the last one"""
        pre = copy(self)
        post = copy(self).H
        if inner is None:
            return post @ pre
        pre.oshape = inner.ishape
        post.ishape = inner.oshape
        return post @ inner @ pre

    @staticmethod
    def fn(linop, x, /):
        return end_pad_with_zeros(x, linop.dim, linop.length)

    @staticmethod
    def adj_fn(linop, y, /):
        return y[linop.slc]

    @staticmethod
    def normal_fn(linop, x, /):
        x[linop.end_slc] = 0.0
        return x

    def split_forward(self, ibatch, obatch):
        if ibatch[self.dim] != slice(None) or obatch[self.dim] != slice(None):
            raise ValueError("Cannot slice a PadEnd linop along truncation dimension")
        return type(self)(self.dim, self.length, self.ishape, self.oshape)

    def split_forward_fn(self, ibatch, obatch, /, data=None):
        if ibatch[self.dim] != slice(None) or obatch[self.dim] != slice(None):
            raise ValueError("Cannot slice a PadEnd linop along truncation dimension")
        return None

    # Linop changes relative size, but can't determine the size itself
    def size(self, dim):
        return None

    def size_fn(self, dim, /, data=None):
        return None

    @staticmethod
    def is_in_slice(a_slice, idx):
        """TODO: unused"""
        if idx < a_slice.start or idx >= a_slice.stop:
            return False
        step = a_slice.step if a_slice.step else 1
        if (idx - a_slice.start) % step == 0:
            return True
        else:
            return False



import torch

from torchlinops.functional import unfold
import sigpy as sp
import cupy as cp
import numpy as np

from utils.benchmark import benchmark
from utils import Indenter, device_ordinal
from utils import from_pytorch, to_pytorch


def main():
    device = torch.device("cuda:0")
    N = 2
    Nx = 100
    Ny = 200

    block_dim = (11, 12)
    stride = (1, 2)
    dtype = torch.complex64
    cp_complex: bool = True

    # mask = torch.zeros(*block_dim, dtype=bool)
    # mask[::2, ::2] = 1

    ### Triton Version ###
    def random_unfold_triton():
        x = torch.randn(N * Nx * Ny, dtype=dtype, device=device).reshape(N, Nx, Ny)
        # x = torch.randn((N, Nx), device=device)
        return unfold(x, block_dim, stride)

    triton_res, _ = benchmark(random_unfold_triton, num_iters=100)
    summarize(triton_res, "triton")

    ### torch.compile Version ###
    # Mildly slower... also takes forever to compile
    # def random_unfold_torch():
    #     x = torch.randn((N, Nx), device=device)
    #     return unfold_torch(x, block_dim, stride)

    # torch_res, _ = benchmark(random_unfold_torch)
    # summarize(torch_res, "torch.compile")

    ### Cupy Version ###
    dev = sp.Device(device_ordinal(device))

    def random_unfold_sp():
        # x = torch.randn((N, Nx), device=device)
        # x = from_pytorch(x)
        xp = dev.xp
        with dev:
            if cp_complex:
                x = xp.random.randn(N, Nx, Ny, dtype=np.float32) + 1j * xp.random.randn(
                    N, Nx, Ny, dtype=np.float32
                )
            else:
                x = xp.random.randn(N, Nx, Ny)
            return sp.array_to_blocks(x, block_dim, stride)

    sp_res, _ = benchmark(random_unfold_sp, num_iters=100)
    summarize(sp_res, "sp")

    # Test correctness
    x = torch.randn((N, Nx, Ny), device=device)
    Bx_triton = unfold(x, block_dim, stride)
    Bx_sp = sp.array_to_blocks(from_pytorch(x), block_dim, stride)
    assert torch.allclose(Bx_triton, to_pytorch(Bx_sp))


def summarize(benchmark_result, name: str):
    with Indenter() as indent:
        print(name)
        with indent:
            indent.print(
                f"Mean Time: {np.mean(benchmark_result['timings_ms']):0.3f} ms"
            )
            indent.print(f"Min Time: {np.min(benchmark_result['timings_ms']):0.3f} ms")
            indent.print(f"Max Time: {np.max(benchmark_result['timings_ms']):0.3f} ms")
            indent.print(f"Memory: {benchmark_result['max_mem_bytes']} bytes")


if __name__ == "__main__":
    main()



from typing import Optional

from copy import copy
from dataclasses import dataclass
from typing import Any, Tuple, List

__all__ = ["ND", "NamedDimension", "ELLIPSES", "ANY", "Dim", "parse_dim_str"]

# Special dim names
ELLIPSES = "..."
ANY = "()"


# Convenience function
def parse_dim_str(s: Optional[str] = None) -> tuple[str]:
    """
    Rules:
    - dim begins with an uppercase letter
    - dim cannot start with a number

    Examples
    --------
    >>> parse_dim_str("ABCD")
    ('A', 'B', 'C', 'D')
    >>> parse_dim_str("NxNyNz")
    ('Nx', 'Ny', 'Nz')
    >>> parse_dim_str("A1B2Kx1Ky2")
    ('A1', 'B2', 'Kx1', 'Ky2')
    >>> parse_dim_str("()A()B")
    ('()', 'A', '()', 'B')
    """
    if s is None or len(s) == 0:
        return tuple()
    parts = []
    current = s[0]
    for char in s[1:]:
        if char == "(":
            parts.append(current)
            current = char
        elif current == "()":
            parts.append(current)
            current = char
        elif char.isupper():
            # New dim
            if char.isdigit():
                raise ValueError(f"Dim cannot start with a digit in dim string {s}")
            parts.append(current)
            current = char
        else:
            current += char
    parts.append(current)
    return tuple(parts)


Dim = parse_dim_str


@dataclass(slots=True, frozen=True)
class NamedDimension:
    name: str
    i: int = 0

    @classmethod
    def infer(cls, dim: Any):
        if isinstance(dim, cls):
            return dim
        if isinstance(dim, str) and len(dim) == 2:
            if dim[1].isdigit():
                return cls(dim[0], int(dim[1]))
        elif dim == ELLIPSES:
            return cls(ELLIPSES)
        elif isinstance(dim, Tuple) or isinstance(dim, List):
            return type(dim)(cls.infer(d) for d in dim)
        return cls(dim)

    def next_unused(self, tup):
        """Get the next dim by index that does not occur in tup"""
        curr = copy(self)
        if self.name == ELLIPSES or self.name == ANY:
            return curr
        while curr in tup:
            curr = curr + 1
        return curr

    def __repr__(self):
        return self.name + ("" if self.i == 0 else str(self.i))

    def __add__(self, k):
        if self.name == ELLIPSES:
            return self
        try:
            return type(self)(self.name, self.i + k)
        except TypeError as e:
            raise TypeError(f"Unsupported NamedDimension add: {self} + {k}", e)

    def __eq__(self, other):
        """Tests for simple string equality"""
        return repr(self) == other

    def __hash__(self):
        """Allow dictionary lookups to work with strings too."""
        return hash(repr(self))


ND = NamedDimension

if __name__ == "__main__":
    import doctest

    doctest.testmod()



from jaxtyping import Shaped, Bool, Integer
from torch import Tensor

import torch

__all__ = [
    "index",
    "index_adjoint",
    "mask2idx",
    "canonicalize_idx",
    "slice2range",
    "ensure_tensor_indexing",
]

IndexOrSlice = Integer[Tensor, "..."] | slice


def index(
    vals: Shaped[Tensor, "..."],
    idx: tuple[IndexOrSlice, ...],
) -> Tensor:
    """
    Parameters
    ----------
    idx : tuple of Tensor or Slice objects
        Index
    """
    if len(vals.shape) < len(idx):
        raise ValueError(
            f"Input value with shape {vals.shape} cannot be indexed with index tensors of length {len(idx)}"
        )
    batch_slc = [slice(None)] * (len(vals.shape) - len(idx))
    idx_batched = (*batch_slc, *idx)
    return vals[idx_batched]


def index_adjoint(
    vals: Shaped[Tensor, "..."],
    idx: tuple[Integer[Tensor, "..."], ...],
    grid_size: tuple[int, ...],
) -> Tensor:
    """
    Parameters
    ----------
    vals :
        Batch size of vals is used to determine batch size of output
    idx : tuple of integer-valued tensors
        Use ensure_tensor_indexing to guarantee this
    grid_size : tuple of ints
        The shape of the output tensor, excluding batch dimensions
    """
    for d, (dim_idx, dim_size) in enumerate(zip(idx, grid_size)):
        if (dim_idx >= dim_size).any() or (dim_idx < -dim_size).any():
            # mask = (dim_idx >= dim_size) | (dim_idx < -dim_size)
            raise IndexError(
                f"Out-of-bounds index for grid of shape {grid_size}: idx[{d}]"
            )

    idx_stacked = torch.stack(idx, dim=0)
    out = multi_grid(vals, idx_stacked, grid_size)
    return out


def mask2idx(mask: Bool[Tensor, "..."]) -> tuple[Integer[Tensor, "..."], ...]:
    """Converts an n-dimensional boolean tensor into an n-tuple of integer tensors
    indexing the True elements of the tensor.

    Parameters
    ----------
    mask : torch.Tensor
        A boolean tensor.

    Returns
    -------
    tuple[torch.Tensor]:
        A tuple of integer tensors indexing the True elements.
    """
    if not mask.dtype == torch.bool:
        raise ValueError(f"Input tensor must be of boolean dtype, but got {mask.dtype}")
    return torch.nonzero(mask, as_tuple=True)


def canonicalize_idx(idx: Integer[Tensor, "..."], dim: int = -1):
    """
    Parameters
    ----------
    idx : [B1... D B2...]
    dim : int
        The dimension of idx to tuple-ify

    Returns
    -------
    D-tuple of [B1... B2...] tensors

    Note: dim is usually 0 or -1

    """
    return tuple(torch.select(idx, dim, i) for i in range(idx.shape[dim]))


### Helper functions
def multi_grid(
    x: torch.Tensor,
    idx: torch.Tensor,
    final_size: tuple,
    raveled: bool = False,
    ravel_dim: int = 0,
):
    """Grid values in x to im_size with indices given in idx
    x: [N... I...]
    idx: [I... ndims] or [I...] if raveled=True
    raveled: Whether the idx still needs to be raveled or not
    ravel_dim: Dimension to ravel over

    Returns:
    Tensor with shape [N... final_size]

    Notes:
    Adjoint of multi_index
    Might need nonnegative indices
    """
    if not raveled:
        if len(final_size) != idx.shape[ravel_dim]:
            raise ValueError(
                f"final_size should be of dimension {idx.shape[-1]} but got {final_size}"
            )
        idx = ravel(idx, final_size, dim=ravel_dim)
    ndim = len(idx.shape)
    if x.shape[-ndim:] != idx.shape:
        raise ValueError(
            f"x and idx should correspond in last {ndim} dimensions but got x: {x.shape} and idx: {idx.shape}"
        )
    x_flat = torch.flatten(x, start_dim=-ndim, end_dim=-1)  # [N... (I...)]
    idx_flat = torch.flatten(idx)

    batch_dims = x_flat.shape[:-1]
    y = torch.zeros(
        (*batch_dims, *final_size), dtype=x_flat.dtype, device=x_flat.device
    )
    y = y.reshape((*batch_dims, -1))
    y = y.index_add_(-1, idx_flat, x_flat)
    y = y.reshape(*batch_dims, *final_size)
    return y


def ravel(x: torch.Tensor, shape: tuple, dim: int):
    """
    x: torch.LongTensor, arbitrary shape,
    shape: Shape of the array that x indexes into
    dim: dimension of x that is the "indexing" dimension

    Returns:
    torch.LongTensor of same shape as x but with indexing dimension removed
    """
    out = 0
    shape_shifted = tuple(shape[1:]) + (1,)
    for s, s_next, i in zip(shape, shape_shifted, range(x.shape[dim])):
        out += torch.select(x, dim, i) % s  # Python does nonnegative modulo
        out *= s_next
    return out


def slice2range(slice_obj: slice, n: int):
    """Convert a slice object to a range object given the array size
    Examples
    --------
    >>> tuple(slice2range(slice(None, None, None), 4))
    (0, 1, 2, 3)
    >>> tuple(slice2range(slice(None, None, -1), 3))
    (2, 1, 0)

    """
    start = (
        slice_obj.start
        if slice_obj.start is not None
        else (0 if slice_obj.step is None or slice_obj.step > 0 else n - 1)
    )
    stop = (
        slice_obj.stop
        if slice_obj.stop is not None
        else (n if slice_obj.step is None or slice_obj.step > 0 else -1)
    )
    step = slice_obj.step if slice_obj.step is not None else 1
    return range(start, stop, step)


def _unsqueeze_last(t: Tensor, n: int):
    """Unsqueeze multiple dimensions at the end of a tensor

    Examples
    --------
    >>> t = torch.arange(3)
    >>> _unsqueeze_last(t, 2).shape
    torch.Size([3, 1, 1])
    """
    return t.view(-1, *((1,) * n))


def ensure_tensor_indexing(
    idx: tuple[IndexOrSlice, ...], tshape: tuple | torch.Size
) -> tuple[Tensor, ...]:
    """Convert any slice()-type indexes to tensor indexes.

    Also broadcasts by appending slice(None) to the front of idx

    Parameters
    ----------
    idx : tuple
        Tuple of torch.Tensor (integer-valued) index tensors or slice() objects
    tshape : torch.Size or tuple
        Target size, should have length greater than or equal to that of idx

    """
    # Prepare idx
    idx = list(idx)
    if len(tshape) < len(idx):
        raise ValueError(f"Cannot broadcast idx {idx} to tshape {tshape}")
    while len(tshape) > len(idx):
        # Insert empty slices until index length matches length of target shape
        idx.insert(0, slice(None))

    # Prepare out
    out = []
    for d, (size, i) in enumerate(zip(tshape, idx)):
        if isinstance(i, Tensor):
            out.append(i)
        elif isinstance(i, slice):
            range_tensor = torch.tensor(slice2range(i, size))
            # Unsqueeze last dimensions
            range_tensor = _unsqueeze_last(range_tensor, len(tshape) - d - 1)
            out.append(range_tensor)
        else:
            raise ValueError(
                f"idx must contain only tensors or slice() objects but got {i}"
            )
    return tuple(out)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from typing import Tuple, Any, Sequence

from ._nameddim import NamedDimension as ND

__all__ = ["fake_dims", "get_nd_shape", "N2K", "K2N"]


def get_nd_shape(im_size, kspace=False):
    if len(im_size) == 1:
        im_dim = ("Kx",) if kspace else ("Nx",)
    elif len(im_size) == 2:
        im_dim = ("Kx", "Ky") if kspace else ("Nx", "Ny")
    elif len(im_size) == 3:
        im_dim = ("Kx", "Ky", "Kz") if kspace else ("Nx", "Ny", "Nz")
    else:
        raise ValueError(f"Image size {im_size} - should have length 2 or 3")
    return im_dim


def fake_dims(letter: str, n: int) -> Tuple:
    """Helper function for generating fake dimension names"""
    return tuple(f"{letter}_{i}" for i in range(n))


def is_spatial_dim(d: ND):
    return "x" in d.name or "y" in d.name or "z" in d.name


def N2K(tup: Tuple[ND]):
    out = []
    for d in tup:
        if is_spatial_dim(d):
            # Flip 'N' to 'K'
            out.append(ND(d.name.replace("N", "K"), d.i))
        else:
            out.append(d)
    return tuple(out)


def K2N(tup: Tuple[ND]):
    out = []
    for d in tup:
        if is_spatial_dim(d):
            # Flip 'N' to 'K'
            out.append(ND(d.name.replace("K", "N"), d.i))
        else:
            out.append(d)
    return tuple(out)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from dataclasses import dataclass, field
from typing import Any, Callable, List
from collections import defaultdict

__all__ = [
    "EventManager",
    "Callback",
]


@dataclass
class Callback:
    name: str
    fn: Callable
    deps: List[str] = field(default_factory=lambda: [])


class EventManager:
    def __init__(self, handlers: dict = None):
        """
        handlers : defaultdict(list)
            defaultdict mapping event names to lists of (callback_name, callback_deps, callback_fn) triples
        """
        self.handlers = defaultdict(list) if handlers is None else handlers
        for event, callbacks in self.handlers.items():
            self.handlers[event] = self.sort_callbacks(callbacks)

    def register_handler(
        self,
        event,
        callback: Callback,
        sort_now: bool = True,
    ):
        self.handlers[event].append(callback)
        if sort_now:
            self.handlers[event] = self.sort_callbacks(self.handlers[event])

    def dispatch(self, event, s: Any):
        for handler in self.handlers[event]:
            s = handler.fn(s)
        return s

    def handlers(self, event):
        return [handler.name for handler in self.handlers[event]]

    @staticmethod
    def sort_callbacks(callbacks):
        """ChatGPT"""
        # Create a mapping from callback name to callback instance
        callback_map = {cb.name: cb for cb in callbacks}

        # Track visited callbacks to avoid cycles and repeated visits
        visited = set()
        # Use a list to act as an ordered stack for the sorted elements
        sorted_callbacks = []

        def dfs(cb: Callback):
            if cb.name in visited:
                return
            visited.add(cb.name)
            # Visit all dependencies first
            for dep_name in cb.deps:
                if dep_name in callback_map:
                    dfs(callback_map[dep_name])
                else:
                    raise ValueError(
                        f"Dependency {dep_name} not found for callback {cb.name}"
                    )
            # Add this callback to the sorted list
            sorted_callbacks.append(cb)

        # Iterate through all callbacks and perform DFS
        for cb in callbacks:
            if cb.name not in visited:
                dfs(cb)

        return sorted_callbacks



import pytest

import torch
from torchlinops.functional import index, index_adjoint


def test_index_2d_bcast():
    idx = (torch.tensor([1, 0, 1])[:, None], torch.tensor([1, 2, 3])[None, :])
    vals = torch.arange(8).reshape(2, 4)
    out = vals[idx]

    out2 = index(vals, idx)
    assert (out == out2).all()


def test_index_2d_bcast_slicing():
    tidx = (torch.tensor([0, 1])[:, None], torch.tensor([1, 2, 3])[None, :])
    vals = torch.arange(8).reshape(2, 4)
    out = vals[tidx]

    idx = (slice(None), slice(1, 4))
    out2 = index(vals, idx)
    assert (out == out2).all()


def test_index_adjoint():
    idx = torch.tensor([1, 3, 2])
    idx = (idx,)
    vals = torch.tensor([[5.0, 4.0, -1.0], [5.0, 4.0, -1.0]])
    oshape = (4,)
    out = index_adjoint(vals, idx, oshape)
    assert (out[0] == torch.tensor([0.0, 5.0, -1.0, 4.0])).all()
    assert (out[0] == out[1]).all()



import pytest
from math import prod

import torch
import torchlinops.functional as F
from torchlinops.functional._unfold.nblocks import get_nblocks

PYTEST_GPU_MARKS = [
    pytest.mark.gpu,
    pytest.mark.skipif(
        not torch.cuda.is_available(), reason="GPU is required but not available"
    ),
]


@pytest.mark.parametrize("dev", ["cpu", pytest.param("cuda", marks=PYTEST_GPU_MARKS)])
@pytest.mark.parametrize("dtype", ["real", "complex"])
@pytest.mark.parametrize(
    "spec",
    [
        "small1d",
        "medium1d",
        "small2d",
        pytest.param("small3d", marks=PYTEST_GPU_MARKS),
    ],
)
def test_a2b(dev, dtype, spec, request):
    spec = request.getfixturevalue(spec)
    device = torch.device(dev)
    dtype = torch.complex64 if dtype == "complex" else torch.float32
    ishape = (*spec["N"], *spec["shape"])

    x = torch.ones(prod(ishape)).to(dtype).reshape(ishape).requires_grad_(True)
    y = F.array_to_blocks(x, spec["block_size"], spec["stride"])
    y.abs().sum().backward()

    gradx = F.blocks_to_array(
        torch.ones_like(y),
        spec["shape"],
        spec["block_size"],
        spec["stride"],
    )
    assert torch.allclose(x.grad, gradx)


@pytest.mark.parametrize("dev", ["cpu", pytest.param("cuda", marks=PYTEST_GPU_MARKS)])
@pytest.mark.parametrize("dtype", ["real", "complex"])
@pytest.mark.parametrize(
    "spec",
    [
        "small1d",
        "medium1d",
        "small2d",
        pytest.param("small3d", marks=PYTEST_GPU_MARKS),
    ],
)
def test_b2a(dev, dtype, spec, request):
    spec = request.getfixturevalue(spec)
    device = torch.device(dev)
    dtype = torch.complex64 if dtype == "complex" else torch.float32
    spec["nblocks"] = get_nblocks(spec["shape"], spec["block_size"], spec["stride"])
    ishape = (*spec["N"], *spec["nblocks"], *spec["block_size"])
    oshape = (*spec["N"], *spec["shape"])

    x = torch.ones(prod(ishape)).to(dtype).reshape(ishape).requires_grad_(True)
    y = F.blocks_to_array(x, spec["shape"], spec["block_size"], spec["stride"])
    y.abs().sum().backward()

    gradx = F.array_to_blocks(
        torch.ones_like(y),
        spec["block_size"],
        spec["stride"],
    )
    assert torch.allclose(x.grad, gradx)



"""Functions for interoperability between cupy and pytorch."""

from torch import Tensor

from warnings import warn
import torch
import numpy as np

try:
    import cupy as cp

    cupy_enabled = True
except ImportError:
    import numpy as cp

    cupy_enabled = False


__all__ = ["to_pytorch", "from_pytorch", "get_device"]


def get_device(arr) -> torch.device:
    """Bootleg version of sigpy's more-robust get_device"""
    if isinstance(arr, np.ndarray):
        return torch.device("cpu")
    elif cupy_enabled and isinstance(arr, cp.ndarray):
        return torch.device(f"cuda:{arr.device.id}")
    elif isinstance(arr, Tensor):
        return arr.device
    else:
        raise ValueError(f"Can't get device of array type: {type(arr)}")


def to_pytorch(array, requires_grad: bool = False):
    """Zero-copy conversion from numpy/cupy array to pytorch tensor.

    For complex array input, returns a tensor with shape + [2],
    where tensor[..., 0] and tensor[..., 1] represent the real
    and imaginary.

    Args:
        array (numpy/cupy array): input.
        requires_grad(bool): Set .requires_grad output tensor
    Returns:
        PyTorch tensor.

    """
    return torch.as_tensor(array).requires_grad_(requires_grad)

    # import torch
    # from torch.utils.dlpack import from_dlpack

    # device = get_device(array)

    # if device.type == "cpu":
    #     tensor = torch.from_numpy(array)
    # else:
    #     tensor = from_dlpack(array.toDlpack())

    # tensor.requires_grad = requires_grad
    # return tensor.contiguous()


def from_pytorch(tensor):  # pragma: no cover
    """Zero-copy conversion from pytorch tensor to numpy/cupy array.

    If iscomplex, then tensor must have the last dimension as 2,
    and the output will be viewed as a complex valued array.

    Args:
        tensor (PyTorch tensor): input.
        iscomplex (bool): whether input represents complex valued tensor.

    Returns:
        Numpy/cupy array.

    """
    device = tensor.device
    if device.type == "cpu":
        output = tensor.detach().contiguous().numpy()
    else:
        if cupy_enabled:
            output = cp.asarray(tensor)
        else:
            raise TypeError(
                "CuPy not installed, but trying to convert GPU PyTorch Tensor."
            )
    return output



from copy import copy
from typing import Union, Optional
from collections import OrderedDict
from collections.abc import Iterable, Sequence

from ._nameddim import ND
from ._nameddimcollection import NamedDimCollection
from ._matching import isequal

__all__ = ["NS", "NamedShape", "NDorStr", "Shape"]

NDorStr = Union[ND, str]
Shape = Iterable[NDorStr]


def NS(ishape: Shape, oshape: Optional[Shape] = None, **additional_shapes):
    """
    If shape is empty, use tuple(), not None
    """
    if ishape is None:
        shape = NamedShape(ishape=("...",), oshape=("...",))
    elif oshape is None:
        if isinstance(ishape, NamedShape):
            return ishape
        shape = NamedShape(ishape=ishape, oshape=ishape)
    else:
        shape = NamedShape(ishape=ishape, oshape=oshape)
    # Option to add extra shapes
    for k, v in additional_shapes.items():
        shape.add(k, v)
    return shape


class NamedShape(NamedDimCollection):
    """A linop shape with input and output dimensions
    Inherit from this to define custom behavior
    - e.g. splitting ishape and oshape into subparts that are linked
    """

    def __init__(self, ishape: Iterable[NDorStr], oshape: Iterable[NDorStr]):
        super().__init__(_ishape=ishape, _oshape=oshape)

    @staticmethod
    def convert(a: Iterable[NDorStr]):
        return list(ND.infer(a))

    def adjoint(self):
        """Return the adjoint shape. Don't call this method directly, but definitely override it"""
        new = type(self)(self.oshape, self.ishape)
        for shape in self.shapes:
            if shape not in ["_ishape", "_oshape"]:
                new.add(shape, self.lookup(shape))
        return new

    def normal(self):
        new_oshape = tuple(d.next_unused(self.ishape) for d in self.ishape)
        new = type(self)(self.ishape, new_oshape)
        for shape in self.shapes:
            if shape not in ["_ishape", "_oshape"]:
                new.add(shape, self.lookup(shape))
        return new

    @property
    def ishape(self) -> tuple[ND]:
        return self._ishape

    @ishape.setter
    def ishape(self, val: Iterable[NDorStr]):
        _ishape = self.convert(val)
        self._ishape = _ishape

    @property
    def oshape(self) -> tuple[ND]:
        return self._oshape

    @oshape.setter
    def oshape(self, val: Iterable[NDorStr]):
        _oshape = self.convert(val)
        self._oshape = _oshape

    @property
    def H(self):
        return self.adjoint()

    @property
    def N(self):
        return self.normal()

    def __repr__(self):
        return f"{self.ishape} -> {self.oshape}"

    def __add__(self, right):
        try:
            _ishape = self.ishape + right.ishape
        except TypeError as e:
            raise TypeError(
                f"Problem combining shapes {self.ishape} + {right.ishape}"
            ) from e
        try:
            _oshape = self.oshape + right.oshape
        except TypeError as e:
            raise TypeError(
                f"Problem combining shapes {self.oshape} + {right.oshape}"
            ) from e
        new = type(self)(ishape=_ishape, oshape=_oshape)
        # Add all the subshapes
        for shape in self.shapes:
            if shape not in ["_ishape", "_oshape"]:
                new.add(shape, self.lookup(shape))
        for shape in right.shapes:
            if shape not in ["_ishape", "_oshape"]:
                new.add(shape, right.lookup(shape))
        return new

    def __radd__(self, left):
        if left is None:
            return self
        return left.__add__(self)

    def __eq__(self, other):
        return isequal(self.ishape, other.ishape) and isequal(self.oshape, other.oshape)



"""Copied from https://github.com/sidward/ppcs
DOI: https://zenodo.org/badge/latestdoi/452385092I
"""

from typing import Callable, Literal

from dataclasses import dataclass
import logging

import numpy as np
from scipy.special import binom
from sympy import (
    symbols,
    simplify,
    stationary_points,
    Interval,
    Poly,
    Float,
    diff,
    integrate,
)

from torchlinops.linops import NamedLinop, Identity

__all__ = ["polynomial_preconditioner"]


def polynomial_preconditioner(
    T: NamedLinop,
    degree: int,
    norm: Literal["l_2", "l_inf", "ifista"] = "l_2",
    lower_eig: float = 0.0,
    upper_eig: float = 1.0,
) -> NamedLinop:
    """Apply polynomial preconditioning to a linop.
    norm : Type of preconditioner.
        - "l_2"    = l_2 optimized polynomial.
        - "l_inf"  = l_inf optimized polynomial.
        - "ifista" = from DOI: 10.1137/140970537.
    degree : int, >= -1
        Degree of polynomial to use
        -1 = no preconditioning
        0+ = preconditioning
    lower_eig, upper_eig : float
        Eigenvalue bounds for coefficient optimization
    Returns
    -------
    P : NamedLinop
        Polynomial preconditioned version of T.
    """
    Id: NamedLinop = Identity()  # Fixing shapes
    if degree < 0:
        return Id

    if norm == "l_2":
        c, _ = l_2_opt(degree, lower_eig, upper_eig)
    elif norm == "l_inf":
        c, _ = l_inf_opt(degree, lower_eig, upper_eig)
    elif norm == "ifista":
        c = ifista_coeffs(degree)
    else:
        raise ValueError(f"Unknown polynomial preconditioning norm option: {norm}")

    def phelper(c) -> NamedLinop:
        if c.size == 1:
            return c[0] * Id
        L = c[0] * Id  # ... -> ...
        R = phelper(c[1:]) @ T @ Id  # ... -> ...
        return L + R  # ... -> ...

    P = phelper(c)
    return P


def l_inf_opt(degree, lower=0, upper=1, verbose=True):
    """
    (coeffs, polyexpr) = l_inf_opt(degree, l=0, L=1, verbose=True)

    Calculate polynomial p(x) that minimizes the supremum of |1 - x p(x)|
    over (l, L).

    Based on Equation 50 of:
       Shewchuk, J. R.
       An introduction to the conjugate gradient method without the agonizing
       pain, Edition 1¼.

    Uses the following package:
      https://github.com/mlazaric/Chebyshev/
      DOI: 10.5281/zenodo.5831845

    Inputs:
      degree (Int): Degree of polynomial to calculate.
      lower (Float): Lower bound of interval.
      upper (Float): Upper bound of interval.
      verbose (Bool): Print information.

    Returns:
      coeffs (Array): Coefficients of optimized polynomial.
      polyexpr (SymPy): Resulting polynomial as a SymPy expression.
    """
    from Chebyshev.chebyshev import polynomial as chebpoly

    assert degree >= 0

    if verbose:
        print("L-infinity optimized polynomial.")
        print("> Degree:   %d" % degree)
        print("> Spectrum: [%0.2f, %0.2f]" % (lower, upper))

    T = chebpoly.get_nth_chebyshev_polynomial(degree + 1)

    y = symbols("y")
    P = T((upper + lower - 2 * y) / (upper - lower))
    P = P / P.subs(y, 0)
    P = simplify((1 - P) / y)

    if verbose:
        print("> Resulting polynomial: %s" % repr(P))

    if degree > 0:
        points = stationary_points(P, y, Interval(lower, upper))
        vals = np.array(
            [P.subs(y, point) for point in points]
            + [P.subs(y, lower)]
            + [P.subs(y, upper)]
        )
        assert np.abs(vals).min() > 1e-8, "Polynomial not injective."

    c = Poly(P).all_coeffs()[::-1] if degree > 0 else (Float(P),)
    return (np.array(c, dtype=np.float32), P)


def l_2_opt(degree, lower=0, upper=1, weight=1, verbose=True):
    """
    (coeffs, polyexpr) = l_2_opt(degree, l=0, L=1, verbose=True)

    Calculate polynomial p(x) that minimizes the following:

    ..math:
      \int_l^l w(x) (1 - x p(x))^2 dx

    To incorporate priors, w(x) can be used to weight regions of the
    interval (l, L) of the expression above.

    Based on:
      Polynomial Preconditioners for Conjugate Gradient Calculations
      Olin G. Johnson, Charles A. Micchelli, and George Paul
      DOI: 10.1137/0720025

    Inputs:
      degree (Int): Degree of polynomial to calculate.
      l (Float): Lower bound of interval.
      L (Float): Upper bound of interval.
      weight (SymPy): Sympy expression to include prior weight.
      verbose (Bool): Print information.

    Returns:
      coeffs (Array): Coefficients of optimized polynomial.
      polyexpr (SymPy): Resulting polynomial as a SymPy expression.
    """
    if verbose:
        print("L-2 optimized polynomial.")
        print("> Degree:   %d" % degree)
        print("> Spectrum: [%0.2f, %0.2f]" % (lower, upper))

    c = symbols("c0:%d" % (degree + 1))
    x = symbols("x")

    p = sum([(c[k] * x**k) for k in range(degree + 1)])
    f = weight * (1 - x * p) ** 2
    J = integrate(f, (x, lower, upper))

    mat = [[0] * (degree + 1) for _ in range(degree + 1)]
    vec = [0] * (degree + 1)

    for edx in range(degree + 1):
        eqn = diff(J, c[edx])
        tmp = eqn.copy()
        # Coefficient index
        for cdx in range(degree + 1):
            mat[edx][cdx] = float(Poly(eqn, c[cdx]).coeffs()[0])
            tmp = tmp.subs(c[cdx], 0)
        vec[edx] = float(-tmp)

    mat = np.array(mat, dtype=np.double)
    vec = np.array(vec, dtype=np.double)
    res = np.array(np.linalg.pinv(mat) @ vec, dtype=np.float32)

    poly = sum([(res[k] * x**k) for k in range(degree + 1)])
    if verbose:
        print("> Resulting polynomial: %s" % repr(poly))

    if degree > 0:
        points = stationary_points(poly, x, Interval(lower, upper))
        vals = np.array(
            [poly.subs(x, point) for point in points]
            + [poly.subs(x, lower)]
            + [poly.subs(x, upper)]
        )
        assert vals.min() > 1e-8, "Polynomial is not positive."

    return (res, poly)


def ifista_coeffs(degree):
    """
    coeffs = ifista_coeffs(degree)

    Returns coefficients from:
      An Improved Fast Iterative Shrinkage Thresholding Algorithm for Image
      Deblurring
      Md. Zulfiquar Ali Bhotto, M. Omair Ahmad, and M. N. S. Swamy
      DOI: 10.1137/140970537

    Inputs:
      degree (Int): Degree of polynomial to calculate.

    Returns:
      coeffs (Array): Coefficients of optimized polynomial.
    """
    c = []
    for k in range(degree + 1):
        c.append(binom(degree + 1, k + 1) * ((-1) ** (k)))
    return np.array(c)



import torch

from torchlinops.utils import inner, is_adjoint

from torchlinops import Dense


def test_dense():
    M, N = 9, 3
    weight = torch.randn(M, N, dtype=torch.complex64)
    weightshape = ("M", "N")
    x = torch.randn(N, dtype=torch.complex64)
    ishape = ("N",)
    y = torch.randn(M, dtype=torch.complex64)
    oshape = ("M",)
    A = Dense(weight, weightshape, ishape, oshape)
    assert is_adjoint(A, x, y)


def test_dense_shapes():
    ishape = ("B", "N")
    oshape = ("B", "M")
    B = 10
    M, N = (3, 7)
    weight = torch.randn(B, M, N)
    weightshape = ("B", "M", "N")
    device = "cpu"
    A = Dense(weight, weightshape, ishape, oshape)

    x = torch.randn(B, N)
    y = torch.randn(B, M)
    AN = A.N
    ANx = AN(x)  # Make sure it runs
    assert AN.ishape == ("B", "N")
    assert AN.oshape == ("B", "N1")
    assert AN.weightshape == ("B", "N1", "N")

    AH = A.H
    AHy = AH(y)  # Make sure it runs
    assert AH.ishape == ("B", "M")
    assert AH.oshape == ("B", "N")
    assert AH.weightshape == ("B", "M", "N")



import itertools

from tqdm import tqdm

__all__ = [
    "ceildiv",
    "batch_iterator",
    "batch_tqdm",
    "dict_product",
]


def ceildiv(dividend, divisor):
    return -(-dividend // divisor)


def batch_iterator(total, batch_size):
    assert total > 0, f"batch_iterator called with {total} elements"
    delim = list(range(0, total, batch_size)) + [total]
    return zip(delim[:-1], delim[1:])


def batch_tqdm(total, batch_size, **tqdm_kwargs):
    iterator = batch_iterator(total, batch_size)
    return tqdm(iterator, total=ceildiv(total, batch_size), **tqdm_kwargs)


def dict_product(input_dict):
    """Generate all possible dictionaries from a dictionary
    mapping keys to iterables

    ChatGPT-4
    """
    # Extract keys and corresponding iterables
    keys, values = zip(*input_dict.items())

    # Generate all combinations using product
    combinations = itertools.product(*values)

    # Create a list of dictionaries for each combination
    result = [dict(zip(keys, combo)) for combo in combinations]

    return result



from collections.abc import Mapping
from copy import copy
from typing import Optional

import torch
import torch.nn as nn
from jaxtyping import Integer
from torch import Tensor
from torchlinops.functional import slice2range
from torchlinops.utils import INDENT

from .add import Add
from .identity import Zero
from .nameddim import ELLIPSES, ND, NS, NDorStr, isequal
from .namedlinop import NamedLinop

__all__ = ["Stack"]


class Stack(NamedLinop):
    """Concatenate some linops along a new dimension

    Linops need not output tensors of the same size, but they should
    output tensors of the same number of dimensions

    Stacking type depends on dimensions provided

    Horizontal stacking
    stacking along an input dimension:

    A B C

    Vertical stacking
    stacking along an output dimension:

    A
    B
    C

    Diagonal stacking:
    stacking along a separate input and output dimensions

    A . .
    . B .
    . . C


    """

    def __init__(
        self,
        *linops,
        idim_and_idx: Optional[tuple[NDorStr, int]] = (None, None),
        odim_and_idx: Optional[tuple[NDorStr, int]] = (None, None),
    ):
        """
        stack_input_dim / stack_output_dim : int
            If not None, inputs will be stacked (rather than concatenated) along the requested
            dimension. idim / odim must NOT be present if the respective stack_* flag is set.
        """
        self._check_linop_compatibility(linops)

        self.idim, self.idim_idx, ishape = self._get_dim_and_idx(
            *idim_and_idx, linops[0].ishape
        )
        self.odim, self.odim_idx, oshape = self._get_dim_and_idx(
            *odim_and_idx, linops[0].oshape
        )

        # Initialize parent class
        super().__init__(NS(ishape, oshape))
        self.linops = nn.ModuleList(list(linops))

    @staticmethod
    def _get_dim_and_idx(dim, idx, shape):
        if dim is not None:
            dim = ND.infer(dim)
            if dim in shape:
                raise ValueError(
                    f"Stack linop attempting to add dim {dim} to shape {shape} but shape already contains {dim}"
                )
            shape = list(shape)
            shape.insert(idx, dim)
        else:
            dim = None
            idx = None
        return dim, idx, shape

    def forward(self, x):
        return self.fn(self, x)

    @staticmethod
    def fn(stack, x):
        return stack._fn(
            x,
            stack.linops,
            stack.idim_idx,
            stack.odim_idx,
        )

    @staticmethod
    def adj_fn(stack, x):
        adj_linops = [linop.H for linop in stack.linops]
        return stack._fn(
            x,
            adj_linops,
            stack.odim_idx,
            stack.idim_idx,
        )

    @staticmethod
    def _fn(x: Tensor, linops, idim_idx, odim_idx):
        """Unifies forward and adjoint functionality for stacked linops"""
        # Split inputs
        if idim_idx is not None:  # Diagonal, Horizontal
            if len(linops) != x.shape[idim_idx]:
                raise ValueError(
                    f"Stack Linop expecting input of size {len(linops)} at dim {idim_idx} got input of size {x.shape} with non-matching stack size {x.shape[idim_idx]}"
                )
            xs = x.tensor_split(len(linops), idim_idx)
            xs = [xi.squeeze(idim_idx) for xi in xs]
        else:  # Vertical
            xs = [x] * len(linops)

        # Compute linop(x) for all xs
        if odim_idx is not None:  # Diagonal, Vertical
            ys = []
            for xi, linop in zip(xs, linops):
                ys.append(linop(xi))
            return torch.stack(ys, dim=odim_idx)

        # Horizontal
        y = 0
        for xi, linop in zip(xs, linops):
            y += linop(xi)
        return y

    @staticmethod
    def normal_fn(stack, x):
        return stack.adj_fn(stack, stack.fn(stack, x))

    def size(self, dim):
        return self.size_fn(dim)

    def size_fn(self, dim, /):
        if dim == self.idim or dim == self.odim:
            return len(self.linops)
        else:
            return self.linops[0].size(dim)

    def split_forward(self, ibatch, obatch):
        """Split stack linop"""
        linop_idxs = set(range(len(self.linops)))
        for i, slc in enumerate(ibatch):
            if i == self.idim_idx:
                linop_idxs &= set(slice2range(slc, len(self.linops)))
        for i, slc in enumerate(obatch):
            if i == self.odim_idx:
                linop_idxs &= set(slice2range(slc, len(self.linops)))

        if len(linop_idxs) == 0:
            # No linops satisfy this slice (diagonal stacking)
            return Zero(self.ishape, self.oshape)
            # elif len(output_linop_idxs) == 1:
            # else:
        linop_idxs = sorted(list(linop_idxs))
        output_linops = []
        # Remove stack dims from slice batch
        if self.idim_idx is not None:
            ibatch = ibatch.copy()
            ibatch.pop(self.idim_idx)
        if self.odim_idx is not None:
            obatch = obatch.copy()
            obatch.pop(self.odim_idx)

        # Slice each sub-linop
        for i in linop_idxs:
            linop = self.linops[i]
            islices = {dim: slc for dim, slc in zip(linop.ishape, ibatch)}
            oslices = {dim: slc for dim, slc in zip(linop.oshape, obatch)}
            slices = strict_update(islices, oslices)
            output_linops.append(linop.split(linop, slices))
        return type(self)(
            *output_linops,
            idim_and_idx=(self.idim, self.idim_idx),
            odim_and_idx=(self.odim, self.odim_idx),
        )

    def split_forward_fn(self, ibatch, obatch, data_list):
        """Split stack linop, making a new stack linop if necessary

        Parameters
        ----------
        data_list : list, same length as linops
            List of data for each linop in this stack linop

        """
        linop_idxs = set(range(len(self.linops)))
        for i, slc in enumerate(ibatch):
            if i == self.idim_idx:
                linop_idxs &= set(slice2range(slc, len(self.linops)))
        for i, slc in enumerate(obatch):
            if i == self.odim_idx:
                linop_idxs &= set(slice2range(slc, len(self.linops)))

        if len(linop_idxs) == 0:
            # No linops satisfy this slice (diagonal stacking)
            return 0.0  # TODO is this ok
        linop_idxs = sorted(list(linop_idxs))
        output_linop_data = []
        # Remove stack dims from slice batch
        if self.idim_idx is not None:
            ibatch = copy(ibatch)
            ibatch.pop(self.idim_idx)
        if self.odim_idx is not None:
            obatch = copy(obatch)
            obatch.pop(self.odim_idx)

        # Slice each sub-linop
        for i in linop_idxs:
            linop = self.linops[i]
            data = data_list[i]
            output_linop_data.append(linop.split_forward_fn(ibatch, obatch, data))
        return output_linop_data

    def adjoint(self):
        adj_linops = [linop.H for linop in self.linops]
        return type(self)(
            *adj_linops,
            idim_and_idx=(self.odim, self.odim_idx),
            odim_and_idx=(self.idim, self.idim_idx),
        )

    def normal(self, inner=None):
        if inner is None:
            if self.idim is None:  # Vertical (inner product)
                # self.odim is not None
                return Add(*(linop.N for linop in self.linops))
            elif self.odim is None:  # Horizontal (outer product)
                # self.idim is not None
                new_idim, new_odim = self._get_new_normal_io_dims(
                    self._shape, self.idim
                )
                rows = []
                new_shape = self.linops[0].shape.N
                for linop_left in self.linops:
                    row = []
                    for linop_right in self.linops:
                        if linop_left == linop_right:
                            new_linop = linop_right.N
                        else:
                            new_linop = linop_left.H @ linop_right
                            new_linop.ishape = new_shape.ishape
                            new_linop.oshape = new_shape.oshape
                        row.append(new_linop)
                    row = type(self)(
                        *row,
                        idim_and_idx=(new_idim, self.idim_idx),
                        odim_and_idx=(None, None),
                    )
                    rows.append(row)
                return type(self)(
                    *rows,
                    idim_and_idx=(None, None),
                    odim_and_idx=(new_odim, self.idim_idx),
                )
            else:  # Diagonal
                # self.idim and self.odim are not None
                diag = []
                new_idim, new_odim = self._get_new_normal_io_dims(
                    self._shape, self.idim
                )
                for linop in self.linops:
                    diag.append(linop.N)
                return type(self)(
                    *diag,
                    idim_and_idx=(new_idim, self.idim_idx),
                    odim_and_idx=(new_odim, self.odim_idx),
                )
        return super().normal(inner)

    @staticmethod
    def _get_new_normal_io_dims(shape, dim) -> tuple:
        new_shape = shape.N
        i = new_shape.ishape.index(dim)
        new_idim = new_shape.ishape[i]
        new_odim = new_shape.oshape[i]
        return new_idim, new_odim

    @staticmethod
    def _check_linop_compatibility(linops: list[NamedLinop]):
        """Ensure linops can actually be concatenated along the requested dimension"""
        target_shape = linops[0].shape
        for linop in linops:
            if not (
                isequal(target_shape.ishape, linop.ishape)
                and isequal(target_shape.oshape, linop.oshape)
            ):
                raise ValueError(
                    f"Incompatible linops being stacked. Target shape: {target_shape} but got linop shape: {linop.shape}"
                )

    def __getitem__(self, idx):
        linops = self.linops[idx]
        if isinstance(linops, NamedLinop):
            return linops
        return type(self)(
            *linops,
            idim_and_idx=(self.idim, self.idim_idx),
            odim_and_idx=(self.odim, self.odim_idx),
        )

    def __len__(self):
        return len(self.linops)

    def __repr__(self):
        output = ""
        output += INDENT.indent(self.repr_name + f"({self._shape}\n")
        with INDENT:
            for linop in self.linops:
                output += repr(linop) + "\n"
            output += INDENT.indent(f"idim = {self.idim}, odim = {self.odim}\n")
        output += INDENT.indent(")")
        return output

def strict_update(d1: Mapping, d2: Mapping) -> Mapping:
    """Strictly updates one dictionary with values from the other.

    Parameters
    ----------
    d1, d2 : Mappings
        The mappings to combine.

    Returns
    -------
    Mapping
        The combined mapping.

    Raises
    ------
    ValueError
        If there are conflicting keys.
    """
    for k, v in d2.items():
        if k in d1 and d1[k] != v:
            raise ValueError(f"Conflict at key '{k}': {d1[k]} != {v}")
    d1.update(d2)
    return d1
